import pytest

from gaphor import UML
from gaphor.diagram.general import CommentItem, CommentLineItem
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.classes import ClassItem, GeneralizationItem
from gaphor.UML.sanitizerservice import SanitizerService


@pytest.fixture(autouse=True)
def sanitizer(event_manager):
    sanitizer = SanitizerService(event_manager)
    yield sanitizer
    sanitizer.shutdown()


@pytest.fixture
def create_item(element_factory, diagram):
    def create(item_cls, subject_cls=None, subject=None):
        """
        Create an item with specified subject.
        """
        if subject_cls is not None:
            subject = element_factory.create(subject_cls)
        item = diagram.create(item_cls, subject=subject)
        diagram.canvas.update()
        return item

    return create


def test_connect_element_with_comments(create_item):
    comment = create_item(CommentItem, UML.Comment)
    line = create_item(CommentLineItem)
    gi = create_item(GeneralizationItem)
    clazz1 = create_item(ClassItem, UML.Class)
    clazz2 = create_item(ClassItem, UML.Class)

    connect(line, line.head, comment)
    connect(line, line.tail, gi)

    assert line.canvas.get_connection(line.tail).connected is gi

    # Now connect generaliztion ends.
    connect(gi, gi.head, clazz1)
    connect(gi, gi.tail, clazz2)

    assert gi.subject in comment.subject.annotatedElement


def test_presentation_delete(create_item, element_factory):
    """
    Remove element if the last instance of an item is deleted.
    """
    klassitem = create_item(ClassItem, UML.Class)
    klass = klassitem.subject

    assert klassitem.subject.presentation[0] is klassitem
    assert klassitem.canvas

    # Delete presentation here:

    klassitem.unlink()

    assert not klassitem.canvas
    assert klass not in element_factory


def test_stereotype_attribute_delete(element_factory):
    """
    This test was applicable to the Sanitizer service, but is now resolved
    by a tweak in the data model (Instances diagram).
    """
    create = element_factory.create
    klass = create(UML.Class)
    stereotype = create(UML.Stereotype)
    st_attr = create(UML.Property)
    stereotype.ownedAttribute = st_attr

    # Apply stereotype to class and create slot
    instspec = UML.model.apply_stereotype(klass, stereotype)
    slot = UML.model.add_slot(instspec, st_attr)

    # Now, what happens if the attribute is deleted:
    assert st_attr in stereotype.ownedMember
    assert slot in instspec.slot

    st_attr.unlink()

    assert [] == list(stereotype.ownedMember)
    assert [] == list(instspec.slot)


def test_extension_disconnect(element_factory):
    create = element_factory.create
    metaklass = create(UML.Class)
    metaklass.name = "Class"
    klass = create(UML.Class)
    stereotype = create(UML.Stereotype)
    st_attr = create(UML.Property)
    stereotype.ownedAttribute = st_attr
    ext = UML.model.create_extension(metaklass, stereotype)

    # Apply stereotype to class and create slot
    instspec = UML.model.apply_stereotype(klass, stereotype)
    UML.model.add_slot(instspec, st_attr)

    assert stereotype in klass.appliedStereotype[:].classifier

    # Causes set event
    del ext.ownedEnd.type

    assert [] == list(klass.appliedStereotype)


def test_extension_deletion(element_factory):
    create = element_factory.create
    metaklass = create(UML.Class)
    metaklass.name = "Class"
    klass = create(UML.Class)
    stereotype = create(UML.Stereotype)
    st_attr = create(UML.Property)
    stereotype.ownedAttribute = st_attr
    ext = UML.model.create_extension(metaklass, stereotype)

    # Apply stereotype to class and create slot
    instspec = UML.model.apply_stereotype(klass, stereotype)
    UML.model.add_slot(instspec, st_attr)

    assert stereotype in klass.appliedStereotype[:].classifier

    ext.unlink()

    assert [] == list(klass.appliedStereotype)


def test_extension_deletion_with_2_metaclasses(element_factory):
    create = element_factory.create
    metaklass = create(UML.Class)
    metaklass.name = "Class"
    metaiface = create(UML.Class)
    metaiface.name = "Interface"
    klass = create(UML.Class)
    iface = create(UML.Interface)
    stereotype = create(UML.Stereotype)
    st_attr = create(UML.Property)
    stereotype.ownedAttribute = st_attr
    ext1 = UML.model.create_extension(metaklass, stereotype)
    UML.model.create_extension(metaiface, stereotype)

    # Apply stereotype to class and create slot
    instspec1 = UML.model.apply_stereotype(klass, stereotype)
    instspec2 = UML.model.apply_stereotype(iface, stereotype)
    UML.model.add_slot(instspec1, st_attr)

    assert stereotype in klass.appliedStereotype[:].classifier
    assert klass in element_factory

    ext1.unlink()

    assert [] == list(klass.appliedStereotype)
    assert klass in element_factory
    assert [instspec2] == list(iface.appliedStereotype)


def test_stereotype_deletion(element_factory):
    create = element_factory.create
    metaklass = create(UML.Class)
    metaklass.name = "Class"
    klass = create(UML.Class)
    stereotype = create(UML.Stereotype)
    st_attr = create(UML.Property)
    stereotype.ownedAttribute = st_attr
    UML.model.create_extension(metaklass, stereotype)

    # Apply stereotype to class and create slot
    instspec = UML.model.apply_stereotype(klass, stereotype)
    UML.model.add_slot(instspec, st_attr)

    assert stereotype in klass.appliedStereotype[:].classifier

    stereotype.unlink()

    assert [] == list(klass.appliedStereotype)



"""
The Core module provides an entry point for Gaphor's core constructs.

An average module should only need to import this module.
"""

from gaphor.action import action
from gaphor.core.eventmanager import event_handler
from gaphor.i18n import gettext
from gaphor.transaction import Transaction, transactional



import pytest
from gaphas.item import Item

from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling.elementfactory import ElementFactory
from gaphor.core.modeling.presentation import Presentation


class StubItem(Presentation, Item):
    pass


@pytest.fixture
def element_factory():
    event_manager = EventManager()
    return ElementFactory(event_manager)



"""
Unittest the storage and parser modules
"""

import re
from io import StringIO

import pytest

from gaphor import UML
from gaphor.application import distribution
from gaphor.diagram.general import CommentItem
from gaphor.storage import storage
from gaphor.storage.xmlwriter import XMLWriter
from gaphor.tests.testcase import TestCase
from gaphor.UML.classes import AssociationItem, ClassItem, InterfaceItem


class PseudoFile:
    def __init__(self):
        self.data = ""

    def write(self, data):
        self.data += data

    def close(self):
        pass


class StorageTestCase(TestCase):
    def test_version_check(self):
        from gaphor.storage.storage import version_lower_than

        assert version_lower_than("0.3.0", (0, 15, 0))
        assert version_lower_than("0", (0, 15, 0))
        assert version_lower_than("0.14", (0, 15, 0))
        assert version_lower_than("0.14.1111", (0, 15, 0))
        assert not version_lower_than("0.15.0", (0, 15, 0))
        assert not version_lower_than("1.33.0", (0, 15, 0))
        assert not version_lower_than("0.15.0b123", (0, 15, 0))
        assert version_lower_than("0.14.0.b1", (0, 15, 0))
        assert not version_lower_than("0.15.b1", (0, 15, 0))
        assert not version_lower_than("0.16.b1", (0, 15, 0))
        assert not version_lower_than("0.15.0.b2", (0, 14, 99))
        assert not version_lower_than("1.2.0rc2-dev0+7fad31a0", (0, 17, 0))

    def test_save_uml(self):
        """Saving gaphor.UML model elements.
        """
        self.element_factory.create(UML.Package)
        self.element_factory.create(UML.Diagram)
        self.element_factory.create(UML.Comment)
        self.element_factory.create(UML.Class)

        out = PseudoFile()
        storage.save(XMLWriter(out), factory=self.element_factory)
        out.close()

        assert "<Package " in out.data
        assert "<Diagram " in out.data
        assert "<Comment " in out.data
        assert "<Class " in out.data

    def test_save_item(self):
        """Save a diagranm item too.
        """
        diagram = self.element_factory.create(UML.Diagram)
        diagram.create(CommentItem, subject=self.element_factory.create(UML.Comment))

        out = PseudoFile()
        storage.save(XMLWriter(out), factory=self.element_factory)
        out.close()

        assert "<Diagram " in out.data
        assert "<Comment " in out.data
        assert "<canvas>" in out.data
        assert ' type="CommentItem"' in out.data, out.data

    def test_load_uml(self):
        """
        Test loading of a freshly saved model.
        """
        self.element_factory.create(UML.Package)
        # diagram is created in TestCase.setUp
        # self.element_factory.create(UML.Diagram)
        self.element_factory.create(UML.Comment)
        self.element_factory.create(UML.Class)

        data = self.save()
        self.load(data)

        assert len(self.element_factory.lselect()) == 4
        assert len(self.element_factory.lselect(UML.Package)) == 1
        # diagram is created in TestCase.setUp
        assert len(self.element_factory.lselect(UML.Diagram)) == 1
        assert len(self.element_factory.lselect(UML.Comment)) == 1
        assert len(self.element_factory.lselect(UML.Class)) == 1

    def test_load_uml_2(self):
        """
        Test loading of a freshly saved model.
        """
        self.element_factory.create(UML.Package)
        self.create(CommentItem, UML.Comment)
        self.create(ClassItem, UML.Class)
        iface = self.create(InterfaceItem, UML.Interface)
        iface.subject.name = "Circus"
        iface.matrix.translate(10, 10)

        data = self.save()
        self.load(data)

        assert len(self.element_factory.lselect()) == 6
        assert len(self.element_factory.lselect(UML.Package)) == 1
        assert len(self.element_factory.lselect(UML.Diagram)) == 1
        d = self.element_factory.lselect(UML.Diagram)[0]
        assert len(self.element_factory.lselect(UML.Comment)) == 1
        assert len(self.element_factory.lselect(UML.Class)) == 1
        assert len(self.element_factory.lselect(UML.Interface)) == 1

        c = self.element_factory.lselect(UML.Class)[0]
        assert c.presentation
        assert c.presentation[0].subject is c
        # assert c.presentation[0].subject.name.startwith('Class')

        iface = self.element_factory.lselect(UML.Interface)[0]
        assert iface.name == "Circus"
        assert len(iface.presentation) == 1
        assert tuple(iface.presentation[0].matrix) == (1, 0, 0, 1, 10, 10), tuple(
            iface.presentation[0].matrix
        )

        # Check load/save of other canvas items.
        assert len(d.canvas.get_all_items()) == 3
        for item in d.canvas.get_all_items():
            assert item.subject, f"No subject for {item}"
        d1 = d.canvas.select(lambda e: isinstance(e, ClassItem))[0]
        assert d1

    def test_load_with_whitespace_name(self):
        difficult_name = "    with space before and after  "
        diagram = self.element_factory.lselect()[0]
        diagram.name = difficult_name
        data = self.save()
        self.load(data)
        elements = self.element_factory.lselect()
        assert len(elements) == 1, elements
        assert elements[0].name == difficult_name, elements[0].name

    @pytest.mark.slow
    def test_load_uml_metamodel(self):
        path = distribution().locate_file("models/UML.gaphor")

        with open(path) as ifile:
            storage.load(
                ifile,
                factory=self.element_factory,
                modeling_language=self.modeling_language,
            )

    def test_save_and_load_model_with_relationships(self):
        self.element_factory.create(UML.Package)
        self.create(CommentItem, UML.Comment)
        self.create(ClassItem, UML.Class)

        a = self.diagram.create(AssociationItem)
        a.handles()[0].pos = (10, 20)
        a.handles()[1].pos = (50, 60)
        assert a.handles()[0].pos.x == 10, a.handles()[0].pos
        assert a.handles()[0].pos.y == 20, a.handles()[0].pos
        assert tuple(a.handles()[1].pos) == (50, 60), a.handles()[1].pos

        data = self.save()
        self.load(data)

        assert len(self.element_factory.lselect()) == 5
        assert len(self.element_factory.lselect(UML.Package)) == 1
        assert len(self.element_factory.lselect(UML.Diagram)) == 1
        d = self.element_factory.lselect(UML.Diagram)[0]
        assert len(self.element_factory.lselect(UML.Comment)) == 1
        assert len(self.element_factory.lselect(UML.Class)) == 1
        assert len(self.element_factory.lselect(UML.Association)) == 0

        # Check load/save of other canvas items.
        assert len(d.canvas.get_all_items()) == 3
        for item in d.canvas.get_all_items():
            if isinstance(item, AssociationItem):
                aa = item
        assert aa
        assert list(map(float, aa.handles()[0].pos)) == [0, 0], aa.handles()[0].pos
        assert list(map(float, aa.handles()[1].pos)) == [40, 40], aa.handles()[1].pos
        d1 = d.canvas.select(lambda e: isinstance(e, ClassItem))[0]
        assert d1

    def test_save_and_load_of_association_with_two_connected_classes(self):
        c1 = self.create(ClassItem, UML.Class)
        c2 = self.create(ClassItem, UML.Class)
        c2.matrix.translate(200, 200)
        self.diagram.canvas.update_matrix(c2)
        assert tuple(self.diagram.canvas.get_matrix_i2c(c2)) == (1, 0, 0, 1, 200, 200)

        a = self.create(AssociationItem)

        self.connect(a, a.head, c1)
        self.connect(a, a.tail, c2)

        self.diagram.canvas.update_now()

        assert a.head.pos.y == 0, a.head.pos
        assert a.tail.pos.x == 10, a.tail.pos
        assert a.tail.pos.y == 200, a.tail.pos
        assert a.subject

        fd = StringIO()
        storage.save(XMLWriter(fd), factory=self.element_factory)
        data = fd.getvalue()
        fd.close()

        old_a_subject_id = a.subject.id

        self.element_factory.flush()
        assert not list(self.element_factory.select())
        fd = StringIO(data)
        storage.load(
            fd, factory=self.element_factory, modeling_language=self.modeling_language
        )
        fd.close()

        diagrams = list(self.kindof(UML.Diagram))
        assert len(diagrams) == 1
        d = diagrams[0]
        a = d.canvas.select(lambda e: isinstance(e, AssociationItem))[0]
        assert a.subject is not None
        assert old_a_subject_id == a.subject.id
        cinfo_head = a.canvas.get_connection(a.head)
        assert cinfo_head.connected is not None
        cinfo_tail = a.canvas.get_connection(a.tail)
        assert cinfo_tail.connected is not None
        assert cinfo_head.connected is not cinfo_tail.connected

    def test_load_and_save_of_a_model(self):
        path = distribution().locate_file("test-models/simple-items.gaphor")

        with open(path, "r") as ifile:
            storage.load(
                ifile,
                factory=self.element_factory,
                modeling_language=self.modeling_language,
            )

        pf = PseudoFile()

        storage.save(XMLWriter(pf), factory=self.element_factory)

        with open(path, "r") as ifile:

            orig = ifile.read()

        copy = pf.data

        expr = re.compile('gaphor-version="[^"]*"')
        orig = expr.sub("%VER%", orig)
        copy = expr.sub("%VER%", copy)

        self.maxDiff = None
        assert copy == orig, "Saved model does not match copy"

    def test_can_not_load_models_older_that_0_17_0(self):

        path = distribution().locate_file("test-models/old-gaphor-version.gaphor")

        def load_old_model():
            with open(path, "r") as ifile:
                storage.load(
                    ifile,
                    factory=self.element_factory,
                    modeling_language=self.modeling_language,
                )

        self.assertRaises(ValueError, load_old_model)



"""
Test extend item connections.
"""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.UML.usecases.extend import ExtendItem
from gaphor.UML.usecases.usecase import UseCaseItem


class ExtendItemTestCase(TestCase):
    def test_use_case_glue(self):
        """Test "extend" gluing to use cases."""

        uc1 = self.create(UseCaseItem, UML.UseCase)
        extend = self.create(ExtendItem)

        glued = self.allow(extend, extend.head, uc1)
        assert glued

    def test_use_case_connect(self):
        """Test connecting "extend" to use cases
        """
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)
        extend = self.create(ExtendItem)

        self.connect(extend, extend.head, uc1)
        assert self.get_connected(extend.head), uc1

        self.connect(extend, extend.tail, uc2)
        assert self.get_connected(extend.tail), uc2

    def test_use_case_reconnect(self):
        """Test reconnecting use cases with "extend"
        """
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)
        uc3 = self.create(UseCaseItem, UML.UseCase)
        extend = self.create(ExtendItem)

        # connect: uc1 -> uc2
        self.connect(extend, extend.head, uc1)
        self.connect(extend, extend.tail, uc2)
        e = extend.subject

        # reconnect: uc1 -> uc2
        self.connect(extend, extend.tail, uc3)

        assert e is extend.subject
        assert extend.subject.extendedCase is uc1.subject
        assert extend.subject.extension is uc3.subject

    def test_use_case_disconnect(self):
        """Test disconnecting "extend" from use cases
        """
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)
        extend = self.create(ExtendItem)

        self.connect(extend, extend.head, uc1)
        self.connect(extend, extend.tail, uc2)

        self.disconnect(extend, extend.head)
        assert self.get_connected(extend.head) is None
        assert extend.subject is None

        self.disconnect(extend, extend.tail)
        assert self.get_connected(extend.tail) is None



import pathlib

import pytest
from gi.repository import GLib

from gaphor.core.eventmanager import EventManager
from gaphor.ui.event import FileLoaded
from gaphor.ui.recentfiles import RecentFiles


class RecentManagerStub:
    def __init__(self):
        self.items = []

    def add_full(self, uri, meta):
        self.items.append(uri)


@pytest.fixture
def event_manager():
    return EventManager()


def test_add_new_recent_file(event_manager):
    recent_manager = RecentManagerStub()
    RecentFiles(event_manager, recent_manager)

    event_manager.handle(FileLoaded(None, "testfile.gaphor"))

    assert len(recent_manager.items) == 1
    assert recent_manager.items[0].startswith("file:///"), recent_manager.items[0]


def test_uri_conversion_with_spaces():
    filename = "/path name/with spaces"
    uri = GLib.filename_to_uri(filename)
    decoded_filename, hostname = GLib.filename_from_uri(uri)
    decoded_posix_filename = pathlib.PurePath(decoded_filename).as_posix()

    assert uri == "file:///path%20name/with%20spaces"
    assert decoded_posix_filename == filename
    assert hostname is None


def test_decode_not_encoded_uri():
    filename = "/path name/with spaces"
    uri = f"file://{filename}"
    decoded_filename, hostname = GLib.filename_from_uri(uri)
    decoded_posix_filename = pathlib.PurePath(decoded_filename).as_posix()

    assert decoded_posix_filename == filename
    assert hostname is None



from gaphor import UML
from gaphor.diagram.tests.fixtures import connect, copy_clear_and_paste
from gaphor.UML.classes import InterfaceItem
from gaphor.UML.components import ComponentItem, ConnectorItem


def test_connector(diagram, element_factory):
    comp = element_factory.create(UML.Component)
    iface = element_factory.create(UML.Interface)

    comp_item = diagram.create(ComponentItem, subject=comp)
    iface_item = diagram.create(InterfaceItem, subject=iface)
    conn_item = diagram.create(ConnectorItem)

    connect(conn_item, conn_item.handles()[0], comp_item)
    connect(conn_item, conn_item.handles()[1], iface_item)

    copy_clear_and_paste({comp_item, iface_item, conn_item}, diagram, element_factory)

    new_conn = element_factory.lselect(UML.Connector)[0]
    new_comp = element_factory.lselect(UML.Component)[0]
    new_iface = element_factory.lselect(UML.Interface)[0]

    assert isinstance(new_conn.presentation[0], ConnectorItem)
    assert new_conn.kind == "assembly"
    assert len(new_conn.end) == 1
    assert new_conn.end[0].role is new_iface
    assert new_conn.end[0].partWithPort in new_comp.ownedPort



import pytest

from gaphor.diagram.text import (
    FontStyle,
    FontWeight,
    Layout,
    TextAlign,
    TextDecoration,
    text_point_at_line,
)


@pytest.mark.parametrize(
    "points,point_at_line",
    [
        [[(0, 0), (20, 20)], (5, -10)],
        [[(0, 0), (-20, 20)], (-15, -10)],
        [[(0, 0), (20, -20)], (5, 5)],
        [[(0, 0), (-20, -20)], (-15, 5)],
    ],
)
def test_point_at_left_top(points, point_at_line):
    x, y = text_point_at_line(points, (10, 5), TextAlign.LEFT)

    assert x == point_at_line[0]
    assert y == point_at_line[1]


@pytest.mark.parametrize(
    "points,point_at_line",
    [
        [[(0, 0), (20, 20)], (5, 25)],
        [[(0, 0), (-20, 20)], (-15, 25)],
        [[(0, 0), (20, -20)], (5, -30)],
        [[(0, 0), (-20, -20)], (-15, -30)],
    ],
)
def test_point_at_right_top(points, point_at_line):
    x, y = text_point_at_line(points, (10, 5), TextAlign.RIGHT)

    assert x == point_at_line[0]
    assert y == point_at_line[1]


@pytest.mark.parametrize(
    "points,point_at_line",
    [
        [[(2.0, 2.0), (22.0, 7.0)], (7, 8.75)],
        [[(2.0, 2.0), (22.0, -3.0)], (7, 3.75)],
        [[(2.0, 2.0), (-18.0, 7.0)], (-13, 8.75)],
        [[(2.0, 2.0), (-18.0, -3.0)], (-13, 3.75)],
        [[(2.0, 2.0), (7.0, 22.0)], (-9.125, 9.5)],
        [[(2.0, 2.0), (7.0, -18.0)], (8.125, -10.5)],
        [[(2.0, 2.0), (-3.0, 22.0)], (3.125, 9.5)],
        [[(2.0, 2.0), (-3.0, -18.0)], (-14.125, -10.5)],
        [[(2.0, 2.0), (22.0, 2.0)], (7, 5.0)],
        [[(2.0, 2.0), (-18.0, 2.0)], (-13, 5.0)],
        [[(2.0, 2.0), (2.0, 22.0)], (5.0, 9.5)],
        [[(2.0, 2.0), (2.0, -18.0)], (5.0, -10.5)],
    ],
)
def test_point_at_center_bottom(points, point_at_line):
    """
    Test aligned at the line text position calculation, horizontal mode
    """
    x, y = text_point_at_line(points, (10, 5), TextAlign.CENTER)

    assert x == pytest.approx(point_at_line[0])
    assert y == pytest.approx(point_at_line[1])


def test_text_with_font_as_dict():
    w, h = Layout(
        "Example",
        {
            "font-family": "sans",
            "font-size": 10,
            "font-style": FontStyle.ITALIC,
            "font-weight": FontWeight.BOLD,
            "text-decoration": TextDecoration.UNDERLINE,
        },
    ).size()
    assert w
    assert h


def test_text_with_font_as_dict_with_values_set_to_none():
    w, h = Layout(
        "Example",
        {
            "font-family": "sans",
            "font-size": 9.5,
            "font-style": None,
            "font-weight": FontWeight.BOLD,
            "text-decoration": TextDecoration.NONE,
        },
    ).size()
    assert w
    assert h


def test_text_with_just_font_as_dict():
    w, h = Layout("Example", {"font-family": "sans", "font-size": 10}).size()
    assert w
    assert h



"""
Test messages.
"""

from gaphor import UML
from gaphor.UML.interactions.message import MessageItem


def test_message_persistence(diagram, element_factory, saver, loader):
    diagram.create(MessageItem, subject=element_factory.create(UML.Message))

    data = saver()
    loader(data)
    new_diagram = next(element_factory.select(UML.Diagram))
    item = new_diagram.canvas.select(MessageItem)[0]

    assert item



"""
Flow item adapter connections.
"""

from typing import Type, Union

from gaphor import UML
from gaphor.diagram.connectors import Connector, UnaryRelationshipConnect
from gaphor.UML.actions.action import (
    AcceptEventActionItem,
    ActionItem,
    SendSignalActionItem,
)
from gaphor.UML.actions.activitynodes import (
    ActivityNodeItem,
    DecisionNodeItem,
    ForkNodeItem,
)
from gaphor.UML.actions.flow import FlowItem
from gaphor.UML.actions.objectnode import ObjectNodeItem


class FlowConnect(UnaryRelationshipConnect):
    """
    Connect FlowItem and Action/ObjectNode, initial/final nodes.
    """

    line: FlowItem

    def allow(self, handle, port):
        line = self.line
        subject = self.element.subject

        if (
            handle is line.head
            and isinstance(subject, UML.FinalNode)
            or handle is line.tail
            and isinstance(subject, UML.InitialNode)
        ):
            return None

        return super().allow(handle, port)

    def reconnect(self, handle, port):
        line = self.line
        old_flow = line.subject
        # Secure properties before old_flow is removed:
        name = old_flow.name
        guard_value = old_flow.guard
        self.connect_subject(handle)
        relation = line.subject
        if old_flow:
            relation.name = name
            if guard_value:
                relation.guard = guard_value
            # old_flow.unlink()

    def connect_subject(self, handle):
        line = self.line

        # TODO: connect opposite side again (in case it's a join/fork or
        #       decision/merge node)
        c1 = self.get_connected(line.head)
        c2 = self.get_connected(line.tail)
        if isinstance(c1, ObjectNodeItem) or isinstance(c2, ObjectNodeItem):
            relation: UML.ActivityEdge = self.relationship_or_new(
                UML.ObjectFlow, UML.ObjectFlow.source, UML.ObjectFlow.target
            )
        else:
            relation = self.relationship_or_new(
                UML.ControlFlow, UML.ControlFlow.source, UML.ControlFlow.target
            )
        line.subject = relation
        opposite = line.opposite(handle)
        otc = self.get_connected(opposite)
        if opposite and isinstance(otc, (ForkNodeItem, DecisionNodeItem)):
            adapter = Connector(otc, line)
            adapter.combine_nodes()

    def disconnect_subject(self, handle):
        super().disconnect_subject(handle)
        line = self.line
        opposite = line.opposite(handle)
        otc = self.get_connected(opposite)
        if opposite and isinstance(otc, (ForkNodeItem, DecisionNodeItem)):
            adapter = Connector(otc, line)
            adapter.decombine_nodes()


Connector.register(ActionItem, FlowItem)(FlowConnect)
Connector.register(ActivityNodeItem, FlowItem)(FlowConnect)
Connector.register(ObjectNodeItem, FlowItem)(FlowConnect)
Connector.register(SendSignalActionItem, FlowItem)(FlowConnect)
Connector.register(AcceptEventActionItem, FlowItem)(FlowConnect)


class FlowForkDecisionNodeConnect(FlowConnect):
    """
    Abstract class with common behaviour for Fork/Join node and
    Decision/Merge node.
    """

    element: Union[ForkNodeItem, DecisionNodeItem]
    fork_node_cls: Type[UML.ControlNode]
    join_node_cls: Type[UML.ControlNode]

    def allow(self, handle, port):
        # No cyclic connect is possible on a Flow/Decision node:
        head, tail = self.line.head, self.line.tail
        subject = self.element.subject

        hct = self.get_connected(head)
        tct = self.get_connected(tail)
        if (
            handle is head
            and tct
            and tct.subject is subject
            or handle is tail
            and hct
            and hct.subject is subject
        ):
            return None

        return super().allow(handle, port)

    def combine_nodes(self):
        """
        Combine join/fork or decision/merge nodes into one diagram item.
        """
        fork_node_cls = self.fork_node_cls
        join_node_cls = self.join_node_cls
        element = self.element
        subject = element.subject
        if len(subject.incoming) > 1 and len(subject.outgoing) < 2:
            UML.model.swap_element(subject, join_node_cls)
            element.request_update()
        elif len(subject.incoming) < 2 and len(subject.outgoing) > 1:
            UML.model.swap_element(subject, fork_node_cls)
            element.request_update()
        elif (
            not element.combined
            and len(subject.incoming) > 1
            and len(subject.outgoing) > 1
        ):
            join_node = subject

            # determine flow class:
            if [f for f in join_node.incoming if isinstance(f, UML.ObjectFlow)]:
                flow_class: Type[UML.ActivityEdge] = UML.ObjectFlow
            else:
                flow_class = UML.ControlFlow

            UML.model.swap_element(join_node, join_node_cls)
            fork_node: UML.ControlNode = element.model.create(fork_node_cls)
            for flow in list(join_node.outgoing):
                flow.source = fork_node
            flow = element.model.create(flow_class)
            flow.source = join_node
            flow.target = fork_node

            element.combined = fork_node

    def decombine_nodes(self):
        """
        Decombine join/fork or decision/merge nodes.
        """
        element = self.element
        if element.combined:
            join_node = element.subject
            cflow = join_node.outgoing[0]  # combining flow
            fork_node = cflow.target
            assert fork_node is element.combined
            join_node_cls = self.join_node_cls
            assert isinstance(join_node, join_node_cls)
            fork_node_cls = self.fork_node_cls
            assert isinstance(fork_node, fork_node_cls)

            if len(join_node.incoming) < 2 or len(fork_node.outgoing) < 2:
                # Move all outgoing edges to the first node (the join node):
                for f in list(fork_node.outgoing):
                    f.source = join_node
                cflow.unlink()
                fork_node.unlink()

                # swap subject to fork node if outgoing > 1
                if len(join_node.outgoing) > 1:
                    assert len(join_node.incoming) < 2
                    UML.model.swap_element(join_node, fork_node_cls)
                element.combined = None

    def connect_subject(self, handle):
        """
        In addition to a subject connect, the subject of the element may
        be changed.
        For readability, parameters are named after the classes used by
        Join/Fork nodes.
        """
        super().connect_subject(handle)

        # Switch class for self.element Join/Fork depending on the number
        # of incoming/outgoing edges.
        self.combine_nodes()

    def disconnect_subject(self, handle):
        super().disconnect_subject(handle)
        if self.element.combined:
            self.decombine_nodes()


@Connector.register(ForkNodeItem, FlowItem)
class FlowForkNodeConnect(FlowForkDecisionNodeConnect):
    """Connect Flow to a ForkNode."""

    fork_node_cls = UML.ForkNode
    join_node_cls = UML.JoinNode


@Connector.register(DecisionNodeItem, FlowItem)
class FlowDecisionNodeConnect(FlowForkDecisionNodeConnect):
    """Connect Flow to a DecisionNode."""

    fork_node_cls = UML.DecisionNode
    join_node_cls = UML.MergeNode



import logging

from gi.repository import Gdk

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.ui.abc import UIComponent
from gaphor.ui.event import DiagramOpened
from gaphor.UML.classes import AssociationItem, ClassItem

logging.basicConfig(level=logging.DEBUG)


class DiagramItemConnectorTestCase(TestCase):
    services = TestCase.services + [
        "main_window",
        "properties",
        "namespace",
        "diagrams",
        "toolbox",
        "export_menu",
        "tools_menu",
        "elementeditor",
    ]

    def setUp(self):
        super().setUp()
        self.component_registry = self.get_service("component_registry")
        self.event_manager = self.get_service("event_manager")
        mw = self.get_service("main_window")
        mw.open()
        self.main_window = mw
        self.event_manager.handle(DiagramOpened(self.diagram))

    def test_item_reconnect(self):
        # Setting the stage:
        ci1 = self.create(ClassItem, UML.Class)
        ci2 = self.create(ClassItem, UML.Class)
        a = self.create(AssociationItem)

        self.connect(a, a.head, ci1)
        self.connect(a, a.tail, ci2)

        assert a.subject
        assert a.head_end.subject
        assert a.tail_end.subject

        the_association = a.subject

        # The act: perform button press event and button release
        view = self.component_registry.get(UIComponent, "diagrams").get_current_view()

        assert self.diagram.canvas is view.canvas

        p = view.get_matrix_i2v(a).transform_point(*a.head.pos)

        event = Gdk.Event()
        event.x, event.y, event.type, event.state = (
            p[0],
            p[1],
            Gdk.EventType.BUTTON_PRESS,
            0,
        )

        view.do_event(event)

        assert the_association is a.subject

        event = Gdk.Event()
        event.x, event.y, event.type, event.state = (
            p[0],
            p[1],
            Gdk.EventType.BUTTON_RELEASE,
            0,
        )

        view.do_event(event)

        assert the_association is a.subject



"""The action definition for the UML toolbox."""


from gaphas.item import SE

from gaphor import UML, diagram
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolboxDefinition, ToolDef
from gaphor.diagram.diagramtools import PlacementTool
from gaphor.UML import diagramitems


def namespace_config(new_item):
    subject = new_item.subject
    diagram = new_item.diagram
    subject.package = diagram.namespace
    subject.name = f"New{type(subject).__name__}"


def initial_pseudostate_config(new_item):
    new_item.subject.kind = "initial"


def history_pseudostate_config(new_item):
    new_item.subject.kind = "shallowHistory"


def metaclass_config(new_item):
    namespace_config(new_item)
    new_item.subject.name = "Class"


# Actions: ((section (name, label, icon_name, shortcut)), ...)
uml_toolbox_actions: ToolboxDefinition = (
    (
        gettext("General"),
        (
            ToolDef(
                "toolbox-pointer",
                gettext("Pointer"),
                "gaphor-pointer-symbolic",
                "Escape",
                item_factory=None,
            ),
            ToolDef(
                "toolbox-line",
                gettext("Line"),
                "gaphor-line-symbolic",
                "l",
                PlacementTool.new_item_factory(diagram.general.Line),
            ),
            ToolDef(
                "toolbox-box",
                gettext("Box"),
                "gaphor-box-symbolic",
                "b",
                PlacementTool.new_item_factory(diagram.general.Box),
                SE,
            ),
            ToolDef(
                "toolbox-ellipse",
                gettext("Ellipse"),
                "gaphor-ellipse-symbolic",
                "e",
                PlacementTool.new_item_factory(diagram.general.Ellipse),
                SE,
            ),
            ToolDef(
                "toolbox-comment",
                gettext("Comment"),
                "gaphor-comment-symbolic",
                "k",
                PlacementTool.new_item_factory(
                    diagram.general.CommentItem, UML.Comment
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-comment-line",
                gettext("Comment line"),
                "gaphor-comment-line-symbolic",
                "<Shift>K",
                PlacementTool.new_item_factory(diagram.general.CommentLineItem),
            ),
        ),
    ),
    (
        gettext("Classes"),
        (
            ToolDef(
                "toolbox-class",
                gettext("Class"),
                "gaphor-class-symbolic",
                "c",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ClassItem, UML.Class, config_func=namespace_config
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-interface",
                gettext("Interface"),
                "gaphor-interface-symbolic",
                "i",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.InterfaceItem,
                    UML.Interface,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-package",
                gettext("Package"),
                "gaphor-package-symbolic",
                "p",
                PlacementTool.new_item_factory(
                    diagramitems.PackageItem, UML.Package, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-composite-association",
                gettext("Composite Association"),
                "gaphor-composite-association-symbolic",
                "<Shift>Z",
                PlacementTool.new_item_factory(diagramitems.AssociationItem),
            ),
            ToolDef(
                "toolbox-shared-association",
                gettext("Shared Association"),
                "gaphor-shared-association-symbolic",
                "<Shift>Q",
                PlacementTool.new_item_factory(diagramitems.AssociationItem),
            ),
            ToolDef(
                "toolbox-association",
                gettext("Association"),
                "gaphor-association-symbolic",
                "<Shift>A",
                PlacementTool.new_item_factory(diagramitems.AssociationItem),
            ),
            ToolDef(
                "toolbox-dependency",
                gettext("Dependency"),
                "gaphor-dependency-symbolic",
                "<Shift>D",
                PlacementTool.new_item_factory(diagramitems.DependencyItem),
            ),
            ToolDef(
                "toolbox-generalization",
                gettext("Generalization"),
                "gaphor-generalization-symbolic",
                "<Shift>G",
                PlacementTool.new_item_factory(diagramitems.GeneralizationItem),
            ),
            ToolDef(
                "toolbox-implementation",
                gettext("Implementation"),
                "gaphor-implementation-symbolic",
                "<Shift>I",
                PlacementTool.new_item_factory(diagramitems.ImplementationItem),
            ),
        ),
    ),
    (
        gettext("Components"),
        (
            ToolDef(
                "toolbox-component",
                gettext("Component"),
                "gaphor-component-symbolic",
                "o",
                PlacementTool.new_item_factory(
                    diagramitems.ComponentItem,
                    UML.Component,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-artifact",
                gettext("Artifact"),
                "gaphor-artifact-symbolic",
                "h",
                PlacementTool.new_item_factory(
                    diagramitems.ArtifactItem,
                    UML.Artifact,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-node",
                gettext("Node"),
                "gaphor-node-symbolic",
                "n",
                PlacementTool.new_item_factory(
                    diagramitems.NodeItem, UML.Node, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-device",
                gettext("Device"),
                "gaphor-device-symbolic",
                "d",
                PlacementTool.new_item_factory(
                    diagramitems.NodeItem, UML.Device, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
        ),
    ),
    (
        gettext("Actions"),
        (
            ToolDef(
                "toolbox-action",
                gettext("Action"),
                "gaphor-action-symbolic",
                "a",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ActionItem, UML.Action, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-initial-node",
                gettext("Initial node"),
                "gaphor-initial-node-symbolic",
                "j",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.InitialNodeItem, UML.InitialNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-activity-final-node",
                gettext("Activity final node"),
                "gaphor-activity-final-node-symbolic",
                "f",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ActivityFinalNodeItem, UML.ActivityFinalNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-flow-final-node",
                gettext("Flow final node"),
                "gaphor-flow-final-node-symbolic",
                "w",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.FlowFinalNodeItem, UML.FlowFinalNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-decision-node",
                gettext("Decision/merge node"),
                "gaphor-decision-node-symbolic",
                "g",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.DecisionNodeItem, UML.DecisionNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-fork-node",
                gettext("Fork/join node"),
                "gaphor-fork-node-symbolic",
                "<Shift>R",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ForkNodeItem, UML.JoinNode
                ),
                handle_index=1,
            ),
            ToolDef(
                "toolbox-object-node",
                gettext("Object node"),
                "gaphor-object-node-symbolic",
                "<Shift>O",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ObjectNodeItem,
                    UML.ObjectNode,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-partition",
                gettext("Partition"),
                "gaphor-partition-symbolic",
                "<Shift>P",
                item_factory=PlacementTool.new_item_factory(diagramitems.PartitionItem),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-flow",
                gettext("Control/object flow"),
                "gaphor-control-flow-symbolic",
                "<Shift>F",
                item_factory=PlacementTool.new_item_factory(diagramitems.FlowItem),
            ),
            ToolDef(
                "toolbox-send-signal-action",
                gettext("Send signal action"),
                "gaphor-send-signal-action-symbolic",
                None,
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.SendSignalActionItem,
                    UML.SendSignalAction,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-accept-event-action",
                gettext("Accept event action"),
                "gaphor-accept-event-action-symbolic",
                None,
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.AcceptEventActionItem,
                    UML.AcceptEventAction,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
        ),
    ),
    (
        gettext("Interactions"),
        (
            ToolDef(
                "toolbox-lifeline",
                gettext("Lifeline"),
                "gaphor-lifeline-symbolic",
                "v",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.LifelineItem,
                    UML.Lifeline,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-message",
                gettext("Message"),
                "gaphor-message-symbolic",
                "M",
                item_factory=PlacementTool.new_item_factory(diagramitems.MessageItem),
            ),
            ToolDef(
                "toolbox-execution-specification",
                gettext("Execution Specification"),
                "gaphor-execution-specification-symbolic",
                None,
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ExecutionSpecificationItem
                ),
                handle_index=0,
            ),
            ToolDef(
                "toolbox-interaction",
                gettext("Interaction"),
                "gaphor-interaction-symbolic",
                "<Shift>N",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.InteractionItem,
                    UML.Interaction,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
        ),
    ),
    (
        gettext("States"),
        (
            ToolDef(
                "toolbox-state",
                gettext("State"),
                "gaphor-state-symbolic",
                "s",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.StateItem, UML.State, config_func=namespace_config
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-initial-pseudostate",
                gettext("Initial Pseudostate"),
                "gaphor-initial-pseudostate-symbolic",
                "<Shift>S",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.PseudostateItem,
                    UML.Pseudostate,
                    initial_pseudostate_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-final-state",
                gettext("Final State"),
                "gaphor-final-state-symbolic",
                "x",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.FinalStateItem, UML.FinalState
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-history-pseudostate",
                gettext("History Pseudostate"),
                "gaphor-pseudostate-symbolic",
                "q",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.PseudostateItem,
                    UML.Pseudostate,
                    history_pseudostate_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-transition",
                gettext("Transition"),
                "gaphor-transition-symbolic",
                "<Shift>T",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.TransitionItem
                ),
            ),
        ),
    ),
    (
        gettext("Use Cases"),
        (
            ToolDef(
                "toolbox-use-case",
                gettext("Use case"),
                "gaphor-use-case-symbolic",
                "u",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.UseCaseItem, UML.UseCase, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-actor",
                gettext("Actor"),
                "gaphor-actor-symbolic",
                "t",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ActorItem, UML.Actor, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-use-case-association",
                gettext("Association"),
                "gaphor-association-symbolic",
                "<Shift>B",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.AssociationItem
                ),
            ),
            ToolDef(
                "toolbox-include",
                gettext("Include"),
                "gaphor-include-symbolic",
                "<Shift>U",
                item_factory=PlacementTool.new_item_factory(diagramitems.IncludeItem),
            ),
            ToolDef(
                "toolbox-extend",
                gettext("Extend"),
                "gaphor-extend-symbolic",
                "<Shift>X",
                item_factory=PlacementTool.new_item_factory(diagramitems.ExtendItem),
            ),
        ),
    ),
    (
        gettext("Profiles"),
        (
            ToolDef(
                "toolbox-profile",
                gettext("Profile"),
                "gaphor-profile-symbolic",
                "r",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.PackageItem, UML.Profile, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-metaclass",
                gettext("Metaclass"),
                "gaphor-metaclass-symbolic",
                "m",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ClassItem, UML.Class, config_func=metaclass_config
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-stereotype",
                gettext("Stereotype"),
                "gaphor-stereotype-symbolic",
                "z",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.ClassItem,
                    UML.Stereotype,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-extension",
                gettext("Extension"),
                "gaphor-extension-symbolic",
                "<Shift>E",
                item_factory=PlacementTool.new_item_factory(diagramitems.ExtensionItem),
            ),
            ToolDef(
                "toolbox-import",
                gettext("Import"),
                "gaphor-import-symbolic",
                "<Shift>M",
                item_factory=PlacementTool.new_item_factory(
                    diagramitems.PackageImportItem
                ),
            ),
        ),
    ),
)



"""
Event Manager.
"""


from generic.event import Event, Handler
from generic.event import Manager as _Manager

from gaphor.abc import Service


def event_handler(*event_types):
    """
    Mark a function/method as an event handler for a particular type of event.
    """

    def wrapper(func):
        func.__event_types__ = event_types
        return func

    return wrapper


class EventManager(Service):
    """
    The Event Manager.
    """

    def __init__(self) -> None:
        self._events = _Manager()

    def shutdown(self) -> None:
        pass

    def subscribe(self, handler: Handler) -> None:
        """
        Register a handler. Handlers are triggered (executed) when specific
        events are emitted through the handle() method.
        """
        event_types = getattr(handler, "__event_types__", None)
        if not event_types:
            raise Exception(f"No event types provided for function {handler}")

        for et in event_types:
            self._events.subscribe(handler, et)

    def unsubscribe(self, handler: Handler) -> None:
        """
        Unregister a previously registered handler.
        """
        event_types = getattr(handler, "__event_types__", None)
        if not event_types:
            raise Exception(f"No event types provided for function {handler}")

        for et in event_types:
            self._events.unsubscribe(handler, et)

    def handle(self, *events: Event) -> None:
        """
        Send event notifications to registered handlers.
        """
        for e in events:
            self._events.handle(e)



"""
Test handle tool functionality.
"""

import pytest
from gaphas.aspect import ConnectionSink
from gaphas.aspect import Connector as ConnectorAspect
from gi.repository import Gtk

from gaphor import UML
from gaphor.diagram.connectors import Connector
from gaphor.diagram.diagramtools import ConnectHandleTool, DiagramItemConnector
from gaphor.diagram.general.comment import CommentItem
from gaphor.diagram.general.commentline import CommentLineItem
from gaphor.ui.event import DiagramOpened
from gaphor.ui.mainwindow import Diagrams
from gaphor.UML.modelinglanguage import UMLModelingLanguage
from gaphor.UML.usecases.actor import ActorItem


@pytest.fixture
def diagrams(event_manager, element_factory, properties):
    window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
    diagrams = Diagrams(
        event_manager, element_factory, properties, UMLModelingLanguage()
    )
    window.add(diagrams.open())
    window.show()
    yield diagrams
    diagrams.close()


@pytest.fixture
def comment(element_factory, diagram):
    return diagram.create(CommentItem, subject=element_factory.create(UML.Comment))


@pytest.fixture
def commentline(diagram):
    return diagram.create(CommentLineItem)


def test_aspect_type(commentline):
    aspect = ConnectorAspect(commentline, commentline.handles()[0])
    assert isinstance(aspect, DiagramItemConnector)


def test_query(comment, commentline):
    assert Connector(comment, commentline)


def test_allow(commentline, comment):
    aspect = ConnectorAspect(commentline, commentline.handles()[0])
    assert aspect.item is commentline
    assert aspect.handle is commentline.handles()[0]

    sink = ConnectionSink(comment, comment.ports()[0])
    assert aspect.allow(sink)


def test_connect(diagram, comment, commentline):
    sink = ConnectionSink(comment, comment.ports()[0])
    aspect = ConnectorAspect(commentline, commentline.handles()[0])
    aspect.connect(sink)
    canvas = diagram.canvas
    cinfo = canvas.get_connection(commentline.handles()[0])
    assert cinfo, cinfo


def current_diagram_view(diagrams):
    """
    Get a view for the current diagram.
    """
    view = diagrams.get_current_view()

    # realize view, forces bounding box recalculation
    while Gtk.events_pending():
        Gtk.main_iteration()

    return view


def test_iconnect(event_manager, element_factory, diagrams):
    """
    Test basic glue functionality using CommentItem and CommentLine
    items.
    """
    diagram = element_factory.create(UML.Diagram)
    event_manager.handle(DiagramOpened(diagram))
    comment = diagram.create(CommentItem, subject=element_factory.create(UML.Comment))

    actor = diagram.create(ActorItem, subject=element_factory.create(UML.Actor))
    actor.matrix.translate(200, 200)
    diagram.canvas.update_matrix(actor)

    line = diagram.create(CommentLineItem)

    view = current_diagram_view(diagrams)
    assert view, "View should be available here"
    comment_bb = view.get_item_bounding_box(comment)

    # select handle:
    handle = line.handles()[-1]
    tool = ConnectHandleTool(view=view)

    tool.grab_handle(line, handle)
    handle.pos = (comment_bb.x, comment_bb.y)
    item = tool.glue(line, handle, handle.pos)
    assert item is not None

    tool.connect(line, handle, handle.pos)
    cinfo = diagram.canvas.get_connection(handle)
    assert cinfo.constraint is not None
    assert cinfo.connected is actor, cinfo.connected

    ConnectorAspect(line, handle).disconnect()

    cinfo = diagram.canvas.get_connection(handle)

    assert cinfo is None


def test_connect_comment_and_actor(event_manager, element_factory, diagrams):
    """Test connect/disconnect on comment and actor using comment-line.
    """
    diagram = element_factory.create(UML.Diagram)
    event_manager.handle(DiagramOpened(diagram))
    comment = diagram.create(CommentItem, subject=element_factory.create(UML.Comment))

    line = diagram.create(CommentLineItem)

    view = current_diagram_view(diagrams)
    assert view, "View should be available here"

    tool = ConnectHandleTool(view)

    # Connect one end to the Comment:
    handle = line.handles()[0]
    tool.grab_handle(line, handle)

    handle.pos = (0, 0)
    sink = tool.glue(line, handle, handle.pos)
    assert sink is not None
    assert sink.item is comment

    tool.connect(line, handle, handle.pos)
    cinfo = diagram.canvas.get_connection(handle)
    assert cinfo is not None, None
    assert cinfo.item is line
    assert cinfo.connected is comment

    # Connect the other end to the Actor:
    actor = diagram.create(ActorItem, subject=element_factory.create(UML.Actor))

    handle = line.handles()[-1]
    tool.grab_handle(line, handle)

    handle.pos = (0, 0)
    sink = tool.glue(line, handle, handle.pos)
    assert sink, f"No sink at {handle.pos}"
    assert sink.item is actor
    tool.connect(line, handle, handle.pos)

    cinfo = view.canvas.get_connection(handle)
    assert cinfo.item is line
    assert cinfo.connected is actor

    # Try to connect far away from any item will only do a full disconnect
    assert len(comment.subject.annotatedElement) == 1, comment.subject.annotatedElement
    assert actor.subject in comment.subject.annotatedElement

    sink = tool.glue(line, handle, (500, 500))
    assert sink is None, sink
    tool.connect(line, handle, (500, 500))

    cinfo = view.canvas.get_connection(handle)
    assert cinfo is None



from gaphor.ui.iconname import get_default_icon_name, get_icon_name
from gaphor.UML import uml as UML


@get_icon_name.register(UML.Class)
def get_name_for_class(element):
    if element.extension:
        return "gaphor-metaclass-symbolic"
    else:
        return get_default_icon_name(element)


@get_icon_name.register(UML.Property)
def get_name_for_property(element):
    if element.association:
        return "gaphor-association-symbolic"
    else:
        return "gaphor-property-symbolic"



"""
Pseudostate diagram items.

See also gaphor.UML.states package description.
"""

from gaphas.util import path_ellipse

from gaphor import UML
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import Box, EditableText, IconBox, Text, stroke
from gaphor.diagram.support import represents
from gaphor.UML.modelfactory import stereotypes_str
from gaphor.UML.states.state import VertexItem


@represents(UML.Pseudostate)
class PseudostateItem(ElementPresentation, VertexItem):
    def __init__(self, id=None, model=None):
        super().__init__(id, model)
        for h in self.handles():
            h.movable = False

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject[Pseudostate].kind", self.update_shapes)

    def update_shapes(self):
        self.shape = IconBox(
            Box(
                draw=draw_history_pseudostate
                if self.subject and self.subject.kind == "shallowHistory"
                else draw_initial_pseudostate
            ),
            Text(text=lambda: stereotypes_str(self.subject),),
            EditableText(text=lambda: self.subject.name or ""),
        )


def draw_initial_pseudostate(box, context, bounding_box):
    """
    Draw initial pseudostate symbol.
    """
    cr = context.cairo
    stroke = context.style["color"]
    if stroke:
        cr.set_source_rgba(*stroke)
    r = 10
    d = r * 2
    path_ellipse(cr, r, r, d, d)
    cr.set_line_width(0.01)
    cr.fill()


def draw_history_pseudostate(box, context, bounding_box):
    cr = context.cairo
    r = 15
    d = r * 2
    path_ellipse(cr, r, r, d, d)
    stroke(context)

    cr.move_to(12, 10)
    cr.line_to(12, 20)
    cr.move_to(18, 10)
    cr.line_to(18, 20)
    cr.move_to(12, 15)
    cr.line_to(18, 15)
    stroke(context, fill=False)



"""
Load and save Gaphor models to Gaphors own XML format.

Three functions are exported:
load(filename)
    load a model from a file
save(filename)
    store the current model in a file
"""

__all__ = ["load", "save"]

import io
import logging
import os.path
import uuid
from functools import partial

import gaphas

from gaphor import application
from gaphor.core.modeling import Diagram, Element
from gaphor.core.modeling.collection import collection
from gaphor.storage import parser

FILE_FORMAT_VERSION = "3.0"
NAMESPACE_MODEL = "http://gaphor.sourceforge.net/model"

log = logging.getLogger(__name__)


def save(writer=None, factory=None, status_queue=None):
    for status in save_generator(writer, factory):
        if status_queue:
            status_queue(status)


def save_generator(writer, factory):
    """
    Save the current model using @writer, which is a
    gaphor.storage.xmlwriter.XMLWriter instance.
    """

    writer.startDocument()
    writer.startPrefixMapping("", NAMESPACE_MODEL)
    writer.startElementNS(
        (NAMESPACE_MODEL, "gaphor"),
        None,
        {
            (NAMESPACE_MODEL, "version"): FILE_FORMAT_VERSION,
            (NAMESPACE_MODEL, "gaphor-version"): application.distribution().version,
        },
    )

    size = factory.size()
    n = 0
    for e in list(factory.values()):
        clazz = e.__class__.__name__
        assert e.id
        writer.startElement(clazz, {"id": str(e.id)})
        e.save(partial(save_element, writer=writer))
        writer.endElement(clazz)

        n += 1
        if n % 25 == 0:
            yield (n * 100) / size

    writer.endElementNS((NAMESPACE_MODEL, "gaphor"), None)
    writer.endPrefixMapping("")
    writer.endDocument()


def save_element(name, value, writer):
    """
    Save attributes and references from items in the gaphor.UML module.
    A value may be a primitive (string, int), a gaphor.core.modeling.collection
    (which contains a list of references to other UML elements) or a
    gaphas.Canvas (which contains canvas items).
    """

    def save_reference(name, value):
        """
        Save a value as a reference to another element in the model.
        This applies to both UML as well as canvas items.
        """
        if value.id:
            writer.startElement(name, {})
            writer.startElement("ref", {"refid": value.id})
            writer.endElement("ref")
            writer.endElement(name)

    def save_collection(name, value):
        """
        Save a list of references.
        """
        if value:
            writer.startElement(name, {})
            writer.startElement("reflist", {})
            for v in value:
                if v.id:
                    writer.startElement("ref", {"refid": v.id})
                    writer.endElement("ref")
            writer.endElement("reflist")
            writer.endElement(name)

    def save_value(name, value):
        """
        Save a value (attribute).
        """
        if value is not None:
            writer.startElement(name, {})
            writer.startElement("val", {})
            if isinstance(value, bool):
                # Write booleans as 0/1.
                writer.characters(str(int(value)))
            else:
                writer.characters(str(value))
            writer.endElement("val")
            writer.endElement(name)

    def save_canvas(value):
        """
        Save attributes and references in a gaphor.diagram.* object.
        The extra attribute reference can be used to force UML
        """
        assert isinstance(value, gaphas.Item)
        writer.startElement("item", {"id": value.id, "type": value.__class__.__name__})
        value.save(save_canvas_item)

        for child in value.canvas.get_children(value):
            save_canvas(child)

        writer.endElement("item")

    def save_canvas_item(name, value):
        """
        Save attributes and references in a gaphor.diagram.* object.
        The extra attribute reference can be used to force UML
        """
        if isinstance(value, collection):
            save_collection(name, value)
        elif isinstance(value, (Element, gaphas.Item)):
            save_reference(name, value)
        else:
            save_value(name, value)

    if isinstance(value, (Element, gaphas.Item)):
        save_reference(name, value)
    elif isinstance(value, collection):
        save_collection(name, value)
    elif isinstance(value, gaphas.Canvas):
        writer.startElement("canvas", {})
        value.save(save_canvas)
        writer.endElement("canvas")
    else:
        save_value(name, value)


def load_elements(elements, factory, modeling_language, gaphor_version="1.0.0"):
    for _ in load_elements_generator(
        elements, factory, modeling_language, gaphor_version
    ):
        pass


def load_elements_generator(elements, factory, modeling_language, gaphor_version):
    """
    Load a file and create a model if possible.
    Exceptions: IOError, ValueError.
    """
    log.debug(f"Loading {len(elements)} elements")

    # The elements are iterated three times:
    size = len(elements) * 3

    def update_status_queue(_n=[0]):
        n = _n[0] = _n[0] + 1
        if n % 30 == 0:
            yield (n * 100) / size

    # First create elements and canvas items in the factory
    # The elements are stored as attribute 'element' on the parser objects:
    yield from _load_elements_and_canvasitems(
        elements, factory, modeling_language, gaphor_version, update_status_queue
    )
    yield from _load_attributes_and_references(elements, update_status_queue)

    for d in factory.lselect(Diagram):
        canvas = d.canvas
        # update_now() is implicitly called when lock is released
        canvas.block_updates = False

    # do a postload:
    for id, elem in list(elements.items()):
        yield from update_status_queue()
        elem.element.postload()


def _load_elements_and_canvasitems(
    elements, factory, modeling_language, gaphor_version, update_status_queue
):
    def create_canvasitems(diagram, canvasitems, parent=None):
        """
        Canvas is a read gaphas.Canvas, items is a list of parser.canvasitem's
        """
        if version_lower_than(gaphor_version, (1, 1, 0)):
            new_canvasitems = upgrade_message_item_to_1_1_0(canvasitems)
            canvasitems.extend(new_canvasitems)
            for item in new_canvasitems:
                elements[item.id] = item

        for item in canvasitems:
            item = upgrade_canvas_item_to_1_0_2(item)
            item = upgrade_canvas_item_to_1_3_0(item)
            if version_lower_than(gaphor_version, (1, 1, 0)):
                item = upgrade_presentation_item_to_1_1_0(item)
            cls = modeling_language.lookup_diagram_item(item.type)
            assert cls, f"No diagram item for type {item.type}"
            item.element = diagram.create_as(cls, item.id, parent=parent)
            create_canvasitems(diagram, item.canvasitems, parent=item.element)

    for id, elem in list(elements.items()):
        yield from update_status_queue()
        if isinstance(elem, parser.element):
            cls = modeling_language.lookup_element(elem.type)
            assert cls, f"Type {elem.type} can not be loaded: no such element"
            elem.element = factory.create_as(cls, id)
            if isinstance(elem.element, Diagram):
                assert elem.canvas
                elem.element.canvas.block_updates = True
                create_canvasitems(elem.element, elem.canvas.canvasitems)
        elif not isinstance(elem, parser.canvasitem):
            raise ValueError(
                f"Item with id {id} and type {type(elem)} can not be instantiated"
            )


def _load_attributes_and_references(elements, update_status_queue):
    for id, elem in list(elements.items()):
        yield from update_status_queue()
        # Ensure that all elements have their element instance ready...
        assert hasattr(elem, "element")

        # load attributes and references:
        for name, value in list(elem.values.items()):
            elem.element.load(name, value)

        for name, refids in list(elem.references.items()):
            if isinstance(refids, list):
                for refid in refids:
                    try:
                        ref = elements[refid]
                    except ValueError:
                        log.exception(
                            f"Invalid ID for reference ({refid}) for element {elem.type}.{name}"
                        )
                    else:
                        elem.element.load(name, ref.element)
            else:
                try:
                    ref = elements[refids]
                except ValueError:
                    log.exception(f"Invalid ID for reference ({refids})")
                else:
                    elem.element.load(name, ref.element)


def load(filename, factory, modeling_language, status_queue=None):
    """
    Load a file and create a model if possible.
    Optionally, a status queue function can be given, to which the
    progress is written (as status_queue(progress)).
    """
    for status in load_generator(filename, factory, modeling_language):
        if status_queue:
            status_queue(status)


def load_generator(filename, factory, modeling_language):
    """
    Load a file and create a model if possible.
    This function is a generator. It will yield values from 0 to 100 (%)
    to indicate its progression.
    """
    if isinstance(filename, io.IOBase):
        log.info("Loading file from file descriptor")
    else:
        log.info(f"Loading file {os.fsdecode(os.path.basename(filename))}")
    try:
        # Use the incremental parser and yield the percentage of the file.
        loader = parser.GaphorLoader()
        for percentage in parser.parse_generator(filename, loader):
            if percentage:
                yield percentage / 2
            else:
                yield percentage
        elements = loader.elements
        gaphor_version = loader.gaphor_version

    except OSError:
        log.exception("File could no be parsed")
        raise

    if version_lower_than(gaphor_version, (0, 17, 0)):
        raise ValueError(
            f"Gaphor model version should be at least 0.17.0 (found {gaphor_version})"
        )

    log.info(f"Read {len(elements)} elements from file")

    factory.flush()
    with factory.block_events():
        try:
            for percentage in load_elements_generator(
                elements, factory, modeling_language, gaphor_version
            ):
                if percentage:
                    yield percentage / 2 + 50
                else:
                    yield percentage
            yield 100
        except Exception as e:
            log.warning(f"file {filename} could not be loaded ({e})")
            raise
    factory.model_ready()


def version_lower_than(gaphor_version, version):
    """
    Only major and minor versions are checked.

    >>> version_lower_than('0.3.0', (0, 15, 0))
    True

    """
    parts = gaphor_version.split(".")

    return tuple(map(int, parts[:2])) < version[:2]


def upgrade_canvas_item_to_1_0_2(item):
    if item.type == "MetaclassItem":
        item.type = "ClassItem"
    elif item.type == "SubsystemItem":
        item.type = "ComponentItem"
    return item


def upgrade_canvas_item_to_1_3_0(item):
    if item.type in ("InitialPseudostateItem", "HistoryPseudostateItem"):
        item.type = "PseudostateItem"
    return item


def upgrade_presentation_item_to_1_1_0(item):
    if "show_stereotypes_attrs" in item.values:
        if item.type in (
            "ClassItem",
            "InterfaceItem",
            "ArtifactItem",
            "ComponentItem",
            "NodeItem",
        ):
            item.values["show_stereotypes"] = item.values["show_stereotypes_attrs"]
        del item.values["show_stereotypes_attrs"]

    if "drawing-style" in item.values:
        if item.type == "InterfaceItem":
            item.values["folded"] = "1" if item.values["drawing-style"] == "3" else "0"
        del item.values["drawing-style"]

    if "show-attributes" in item.values and item.type in ("ClassItem", "InterfaceItem"):
        item.values["show_attributes"] = item.values["show-attributes"]
        del item.values["show-attributes"]

    if "show-operations" in item.values and item.type in ("ClassItem", "InterfaceItem"):
        item.values["show_operations"] = item.values["show-operations"]
        del item.values["show-operations"]

    return item


def clone_canvasitem(item, subject_id):
    assert not item.canvasitems, "Can not clone a canvas item with children"
    assert isinstance(item.references["subject"], str)
    new_item = parser.canvasitem(str(uuid.uuid1()), item.type)
    new_item.values = dict(item.values)
    new_item.references = dict(item.references)
    new_item.references["subject"] = subject_id
    return new_item


def upgrade_message_item_to_1_1_0(canvasitems):
    """
    Create new MessageItem's for each `message` and `inverted` message.
    """
    new_canvasitems = []
    for item in canvasitems:
        # new_canvasitems.append(item)
        if item.type == "MessageItem" and item.references.get("subject"):
            messages = item.references.get("message", [])
            inverted = item.references.get("inverted", [])
            if messages:
                del item.references["message"]
            if inverted:
                del item.references["inverted"]
            for m_id in messages:
                new_item = clone_canvasitem(item, m_id)
                new_canvasitems.append(new_item)
            for m_id in inverted:
                new_item = clone_canvasitem(item, m_id)
                new_canvasitems.append(new_item)
                # todo: invert handles, points will follow on connect
                (
                    new_item.references["head-connection"],
                    new_item.references["tail-connection"],
                ) = (
                    new_item.references["tail-connection"],
                    new_item.references["head-connection"],
                )
    return new_canvasitems



"""Activity partition property page."""

from gaphor.core import transactional
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages, new_builder
from gaphor.UML.actions.partition import PartitionItem


@PropertyPages.register(PartitionItem)
class PartitionPropertyPage(PropertyPageBase):
    """Partition property page."""

    order = 15

    def __init__(self, item):
        self.item = item

    def construct(self):
        item = self.item

        if item.toplevel:
            return

        builder = new_builder("partition-editor")

        external = builder.get_object("external")
        external.set_active(item.subject and item.subject.isExternal)

        builder.connect_signals({"external-changed": (self._on_external_change,)})

        return builder.get_object("partition-editor")

    @transactional
    def _on_external_change(self, button, gparam):
        item = self.item
        if item.subject:
            item.subject.isExternal = button.get_active()
        item.request_update()



"""
Copy / Paste functionality
"""

from typing import Set

from gi.repository import Gdk, Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.core import Transaction, action, event_handler
from gaphor.core.modeling import Presentation
from gaphor.diagram.copypaste import copy, paste
from gaphor.ui.event import DiagramSelectionChanged


class CopyService(Service, ActionProvider):
    """
    Copy/Cut/Paste functionality required a lot of thinking:

    Store a list of DiagramItems that have to be copied in a global
    'copy-buffer'.

    - In order to make copy/paste work, the load/save functions should be
      generalized to allow a subset to be saved/loaded (which is needed
      anyway for exporting/importing stereotype Profiles).
    - How much data should be saved? An example use case is to copy a diagram
      item, remove it (the underlying UML element is removed), and then paste
      the copied item. The diagram should act as if we have placed a copy of
      the removed item on the canvas and make the UML element visible again.
    """

    def __init__(self, event_manager, element_factory, diagrams):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.diagrams = diagrams
        self.copy_buffer: object = None

        event_manager.subscribe(self._update)

        self.clipboard = Gtk.Clipboard.get_default(Gdk.Display.get_default())
        self.clipboard.connect("owner_change", self.on_clipboard_owner_change)
        self.clipboard_semaphore = 0

    def shutdown(self):
        self.copy_buffer = set()
        self.event_manager.unsubscribe(self._update)

    def on_clipboard_owner_change(self, clipboard, event):
        if self.clipboard_semaphore > 0:
            self.clipboard_semaphore -= 1
        else:
            self.copy_buffer = set()
            view = self.diagrams.get_current_view()
            self.update_paste_state(view)

    @event_handler(DiagramSelectionChanged)
    def _update(self, event):
        diagram_view = event.diagram_view
        win_action_group = diagram_view.get_action_group("win")
        if win_action_group:
            win_action_group.lookup_action("edit-copy").set_enabled(
                bool(diagram_view.selected_items)
            )

    def copy(self, items):
        if items:
            self.copy_buffer = copy(items)

    def paste(self, diagram):
        """
        Paste items in the copy-buffer to the diagram
        """
        canvas = diagram.canvas

        with Transaction(self.event_manager):
            # Create new id's that have to be used to create the items:
            new_items: Set[Presentation] = paste(
                self.copy_buffer, diagram, self.element_factory.lookup
            )

            # move pasted items a bit, so user can see result of his action :)
            for item in new_items:
                if canvas.get_parent(item) not in new_items:
                    item.matrix.translate(10, 10)

            canvas.update_matrices(new_items)

        return new_items

    def update_paste_state(self, view):
        win_action_group = view.get_action_group("win")
        if win_action_group:
            win_action_group.lookup_action("edit-paste").set_enabled(
                bool(self.copy_buffer)
            )

    @action(
        name="edit-copy", shortcut="<Primary>c",
    )
    def copy_action(self):
        view = self.diagrams.get_current_view()
        if view.is_focus():
            self.clipboard_semaphore += 1
            self.clipboard.set_text("", -1)
            items = view.selected_items
            self.copy(items)
        self.update_paste_state(view)

    @action(name="edit-cut", shortcut="<Primary>x")
    def cut_action(self):
        view = self.diagrams.get_current_view()
        if view.is_focus():
            self.clipboard_semaphore += 1
            self.clipboard.set_text("", -1)
            items = view.selected_items
            self.copy(items)
            for i in list(items):
                i.unlink()
        self.update_paste_state(view)

    @action(name="edit-paste", shortcut="<Primary>v")
    def paste_action(self):
        view = self.diagrams.get_current_view()
        diagram = self.diagrams.get_current_diagram()
        if not (view and view.is_focus()):
            return

        if not self.copy_buffer:
            return

        new_items = self.paste(diagram)

        view.unselect_all()

        for item in new_items:
            view.select_item(item)

        self.event_manager.handle(DiagramSelectionChanged(view, None, new_items))



from gaphor import UML
from gaphor.diagram.tests.fixtures import connect
from gaphor.UML.interactions.executionspecification import ExecutionSpecificationItem
from gaphor.UML.interactions.interactionsconnect import order_lifeline_covered_by
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.interactions.message import MessageItem


def test_ordering(diagram, element_factory):
    lifeline = diagram.create(
        LifelineItem, subject=element_factory.create(UML.Lifeline)
    )
    lifeline.lifetime.visible = True
    lifeline.lifetime.bottom.pos.y = 1000
    lifetime_top = lifeline.lifetime.top.pos

    exec_spec = diagram.create(ExecutionSpecificationItem)
    message1 = diagram.create(MessageItem)
    message2 = diagram.create(MessageItem)
    message3 = diagram.create(MessageItem)

    message1.head.pos.y = lifetime_top.y + 300
    exec_spec.top.pos.y = lifetime_top.y + 400
    message2.head.pos.y = lifetime_top.y + 500
    exec_spec.bottom.pos.y = lifetime_top.y + 600
    message3.tail.pos.y = lifetime_top.y + 700

    # Add in "random" order
    connect(exec_spec, exec_spec.handles()[0], lifeline, lifeline.lifetime.port)
    connect(message3, message3.tail, lifeline, lifeline.lifetime.port)
    connect(message1, message1.head, lifeline, lifeline.lifetime.port)
    connect(message2, message2.head, exec_spec, exec_spec.ports()[1])

    occurrences = [
        message1.subject.sendEvent,
        exec_spec.subject.start,
        message2.subject.sendEvent,
        exec_spec.subject.finish,
        message3.subject.receiveEvent,
    ]

    order_lifeline_covered_by(lifeline)

    assert list(lifeline.subject.coveredBy) == occurrences



"""The action definition for the SysML toolbox."""

from enum import Enum

from gaphas.item import SE

from gaphor import UML, diagram
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolboxDefinition, ToolDef
from gaphor.diagram.diagramtools import PlacementTool
from gaphor.SysML import diagramitems as sysml_items
from gaphor.SysML import sysml
from gaphor.UML import diagramitems as uml_items
from gaphor.UML.toolbox import namespace_config


class AssociationType(Enum):
    COMPOSITE = "composite"
    SHARED = "shared"


def initial_pseudostate_config(new_item):
    new_item.subject.kind = "initial"


def history_pseudostate_config(new_item):
    new_item.subject.kind = "shallowHistory"


def metaclass_config(new_item):
    namespace_config(new_item)
    new_item.subject.name = "Class"


def create_association(
    assoc_item: uml_items.AssociationItem, association_type: AssociationType
) -> None:
    assoc = assoc_item.subject
    assoc.memberEnd.append(assoc_item.model.create(UML.Property))
    assoc.memberEnd.append(assoc_item.model.create(UML.Property))

    assoc_item.head_end.subject = assoc.memberEnd[0]
    assoc_item.tail_end.subject = assoc.memberEnd[1]

    UML.model.set_navigability(assoc, assoc_item.head_end.subject, True)
    assoc_item.head_end.subject.aggregation = association_type.value


def composite_association_config(assoc_item: uml_items.AssociationItem) -> None:
    create_association(assoc_item, AssociationType.COMPOSITE)


def shared_association_config(assoc_item: uml_items.AssociationItem) -> None:
    create_association(assoc_item, AssociationType.SHARED)


# Actions: ((section (name, label, icon_name, shortcut)), ...)
sysml_toolbox_actions: ToolboxDefinition = (
    (
        gettext("General"),
        (
            ToolDef(
                "toolbox-pointer",
                gettext("Pointer"),
                "gaphor-pointer-symbolic",
                "Escape",
                item_factory=None,
            ),
            ToolDef(
                "toolbox-line",
                gettext("Line"),
                "gaphor-line-symbolic",
                "l",
                PlacementTool.new_item_factory(diagram.general.Line),
            ),
            ToolDef(
                "toolbox-box",
                gettext("Box"),
                "gaphor-box-symbolic",
                "b",
                PlacementTool.new_item_factory(diagram.general.Box),
                SE,
            ),
            ToolDef(
                "toolbox-ellipse",
                gettext("Ellipse"),
                "gaphor-ellipse-symbolic",
                "e",
                PlacementTool.new_item_factory(diagram.general.Ellipse),
                SE,
            ),
            ToolDef(
                "toolbox-comment",
                gettext("Comment"),
                "gaphor-comment-symbolic",
                "k",
                PlacementTool.new_item_factory(
                    diagram.general.CommentItem, UML.Comment
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-comment-line",
                gettext("Comment line"),
                "gaphor-comment-line-symbolic",
                "<Shift>K",
                PlacementTool.new_item_factory(diagram.general.CommentLineItem),
            ),
        ),
    ),
    (
        gettext("Blocks"),
        (
            ToolDef(
                "toolbox-block",
                gettext("Block"),
                "gaphor-block-symbolic",
                "l",
                item_factory=PlacementTool.new_item_factory(
                    sysml_items.BlockItem, sysml.Block, config_func=namespace_config
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-package",
                gettext("Package"),
                "gaphor-package-symbolic",
                "p",
                PlacementTool.new_item_factory(
                    uml_items.PackageItem, UML.Package, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-composite-association",
                gettext("Composite Association"),
                "gaphor-composite-association-symbolic",
                "<Shift>Z",
                PlacementTool.new_item_factory(
                    uml_items.AssociationItem,
                    UML.Association,
                    config_func=composite_association_config,
                ),
            ),
            ToolDef(
                "toolbox-shared-association",
                gettext("Shared Association"),
                "gaphor-shared-association-symbolic",
                "<Shift>Q",
                PlacementTool.new_item_factory(
                    uml_items.AssociationItem,
                    UML.Association,
                    config_func=shared_association_config,
                ),
            ),
            ToolDef(
                "toolbox-association",
                gettext("Association"),
                "gaphor-association-symbolic",
                "<Shift>A",
                PlacementTool.new_item_factory(uml_items.AssociationItem),
            ),
            ToolDef(
                "toolbox-generalization",
                gettext("Generalization"),
                "gaphor-generalization-symbolic",
                "<Shift>G",
                PlacementTool.new_item_factory(uml_items.GeneralizationItem),
            ),
        ),
    ),
    (
        gettext("Internal Blocks"),
        (
            ToolDef(
                "toolbox-connector",
                gettext("Connector"),
                "gaphor-connector-symbolic",
                "<Shift>C",
                PlacementTool.new_item_factory(uml_items.ConnectorItem),
            ),
            ToolDef(
                "toolbox-property",
                gettext("Property"),
                "gaphor-property-symbolic",
                "o",
                PlacementTool.new_item_factory(
                    sysml_items.PropertyItem, UML.Property, config_func=namespace_config
                ),
            ),
            ToolDef(
                "toolbox-proxy-port",
                gettext("Proxy Port"),
                "gaphor-proxyport-symbolic",
                "x",
                PlacementTool.new_item_factory(sysml_items.ProxyPortItem),
            ),
        ),
    ),
    (
        gettext("Requirements"),
        (
            ToolDef(
                "toolbox-requirement",
                gettext("Requirement"),
                "gaphor-requirement-symbolic",
                "r",
                item_factory=PlacementTool.new_item_factory(
                    sysml_items.RequirementItem,
                    sysml.Requirement,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-satisfy-dependency",
                gettext("Satisfy"),
                "gaphor-satisfy-symbolic",
                "<Shift>I",
                PlacementTool.new_item_factory(sysml_items.SatisfyItem),
            ),
            ToolDef(
                "toolbox-derive-reqt-dependency",
                gettext("Derive Reqt"),
                "gaphor-derive-symbolic",
                "<Shift>D",
                PlacementTool.new_item_factory(sysml_items.DeriveReqtItem),
            ),
            ToolDef(
                "toolbox-trace-dependency",
                gettext("Trace"),
                "gaphor-trace-symbolic",
                "<Shift>E",
                PlacementTool.new_item_factory(sysml_items.TraceItem),
            ),
            ToolDef(
                "toolbox-refine-dependency",
                gettext("Refine"),
                "gaphor-refine-symbolic",
                "<Shift>N",
                PlacementTool.new_item_factory(sysml_items.RefineItem),
            ),
            ToolDef(
                "toolbox-verify-dependency",
                gettext("Verify"),
                "gaphor-verify-symbolic",
                "<Shift>V",
                PlacementTool.new_item_factory(sysml_items.VerifyItem),
            ),
        ),
    ),
    (
        gettext("Actions"),
        (
            ToolDef(
                "toolbox-action",
                gettext("Action"),
                "gaphor-action-symbolic",
                "a",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.ActionItem, UML.Action, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-initial-node",
                gettext("Initial node"),
                "gaphor-initial-node-symbolic",
                "j",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.InitialNodeItem, UML.InitialNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-activity-final-node",
                gettext("Activity final node"),
                "gaphor-activity-final-node-symbolic",
                "f",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.ActivityFinalNodeItem, UML.ActivityFinalNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-flow-final-node",
                gettext("Flow final node"),
                "gaphor-flow-final-node-symbolic",
                "w",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.FlowFinalNodeItem, UML.FlowFinalNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-decision-node",
                gettext("Decision/merge node"),
                "gaphor-decision-node-symbolic",
                "g",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.DecisionNodeItem, UML.DecisionNode
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-fork-node",
                gettext("Fork/join node"),
                "gaphor-fork-node-symbolic",
                "<Shift>R",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.ForkNodeItem, UML.JoinNode
                ),
                handle_index=1,
            ),
            ToolDef(
                "toolbox-object-node",
                gettext("Object node"),
                "gaphor-object-node-symbolic",
                "<Shift>O",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.ObjectNodeItem,
                    UML.ObjectNode,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-partition",
                gettext("Partition"),
                "gaphor-partition-symbolic",
                "<Shift>P",
                item_factory=PlacementTool.new_item_factory(uml_items.PartitionItem),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-flow",
                gettext("Control/object flow"),
                "gaphor-control-flow-symbolic",
                "<Shift>F",
                item_factory=PlacementTool.new_item_factory(uml_items.FlowItem),
            ),
            ToolDef(
                "toolbox-send-signal-action",
                gettext("Send signal action"),
                "gaphor-send-signal-action-symbolic",
                None,
                item_factory=PlacementTool.new_item_factory(
                    uml_items.SendSignalActionItem,
                    UML.SendSignalAction,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-accept-event-action",
                gettext("Accept event action"),
                "gaphor-accept-event-action-symbolic",
                None,
                item_factory=PlacementTool.new_item_factory(
                    uml_items.AcceptEventActionItem,
                    UML.AcceptEventAction,
                    config_func=namespace_config,
                ),
                handle_index=SE,
            ),
        ),
    ),
    (
        gettext("States"),
        (
            ToolDef(
                "toolbox-state",
                gettext("State"),
                "gaphor-state-symbolic",
                "s",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.StateItem, UML.State, config_func=namespace_config
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-initial-pseudostate",
                gettext("Initial Pseudostate"),
                "gaphor-initial-pseudostate-symbolic",
                "<Shift>S",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.PseudostateItem,
                    UML.Pseudostate,
                    initial_pseudostate_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-final-state",
                gettext("Final State"),
                "gaphor-final-state-symbolic",
                "x",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.FinalStateItem, UML.FinalState
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-history-pseudostate",
                gettext("History Pseudostate"),
                "gaphor-pseudostate-symbolic",
                "q",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.PseudostateItem,
                    UML.Pseudostate,
                    history_pseudostate_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-transition",
                gettext("Transition"),
                "gaphor-transition-symbolic",
                "<Shift>T",
                item_factory=PlacementTool.new_item_factory(uml_items.TransitionItem),
            ),
        ),
    ),
    (
        gettext("Use Cases"),
        (
            ToolDef(
                "toolbox-use-case",
                gettext("Use case"),
                "gaphor-use-case-symbolic",
                "u",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.UseCaseItem, UML.UseCase, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-actor",
                gettext("Actor"),
                "gaphor-actor-symbolic",
                "t",
                item_factory=PlacementTool.new_item_factory(
                    uml_items.ActorItem, UML.Actor, config_func=namespace_config,
                ),
                handle_index=SE,
            ),
            ToolDef(
                "toolbox-use-case-association",
                gettext("Association"),
                "gaphor-association-symbolic",
                "<Shift>B",
                item_factory=PlacementTool.new_item_factory(uml_items.AssociationItem),
            ),
            ToolDef(
                "toolbox-include",
                gettext("Include"),
                "gaphor-include-symbolic",
                "<Shift>U",
                item_factory=PlacementTool.new_item_factory(uml_items.IncludeItem),
            ),
            ToolDef(
                "toolbox-extend",
                gettext("Extend"),
                "gaphor-extend-symbolic",
                "<Shift>X",
                item_factory=PlacementTool.new_item_factory(uml_items.ExtendItem),
            ),
        ),
    ),
)



import pytest

from gaphor.ui.elementeditor import ElementEditor


class DiagramsStub:
    def get_current_view(self):
        return None


@pytest.fixture
def diagrams():
    return DiagramsStub()


class DummyProperties(dict):
    def set(self, key, val):
        self[key] = val


def test_reopen_of_window(event_manager, element_factory, diagrams):
    properties = DummyProperties()
    editor = ElementEditor(event_manager, element_factory, diagrams, properties)

    editor.open()
    editor.close()
    editor.open()
    editor.close()



import ast
import sys
from collections import OrderedDict


def msg(s):
    print(s, file=sys.stderr)


class Writer:
    def __init__(self, overrides=None):
        self.overrides = overrides
        self.classdefs: OrderedDict = OrderedDict()
        self.features = []

    def write(self, filename, header):
        if filename:
            out = hasattr(filename, "write") and filename or open(filename, "w")
        else:
            out = sys.stdout

        try:
            out.write(header)
            if self.overrides:
                out.write(self.overrides.header)
            for cls in self.classdefs.values():
                if cls[0].endswith(":"):
                    for d in cls:
                        out.write(d)
                        out.write("\n")
                    if len(cls) == 1:
                        out.write("    pass\n")
                else:
                    out.write(cls[0])
                    out.write("\n")

                out.write("\n\n")

            for d in self.features:
                out.write(d)
                out.write("\n")
        finally:
            out.close()

    def add_classdef(self, clazz):
        """
        Write a class definition (class xx(x): pass).
        First the parent classes are examined. After that its own definition
        is written. It is ensured that class definitions are only written
        once.
        """
        if not clazz.written:
            s = ""
            for g in clazz.generalization:
                self.add_classdef(g)
                if s:
                    s += ", "
                s += g["name"]
            override = self.overrides.get_override(clazz["name"])
            if override:
                self.classdefs[clazz["name"]] = [override]
            else:
                line = f"class {clazz['name']}"
                if s:
                    line += f"({s})"
                line += ":"
                self.classdefs[clazz["name"]] = [line]
        clazz.written = True

    def add_comment(self, line):
        self.features.append(f"# {line}")

    def add_property(self, class_name, name, value, type):
        """
        Write a property to the file. If the property is overridden, use the
        overridden value. value is
        free format text.
        """
        full_name = f"{class_name}.{name}"
        if self.overrides.has_override(full_name):
            type = self.overrides.get_type(full_name)
            impl = self.overrides.get_override(full_name)
        else:
            impl = f"{class_name}.{name} = {value}"

        decl = f"    {name}: {type}"
        if decl in self.classdefs[class_name]:
            msg(f"feature '{class_name}.{name}' is already defined; skipping")
            assert (
                impl in self.features
            ), f"The implementations for {class_name}.{name} should match"
        else:
            self.classdefs[class_name].append(decl)
            self.features.append(impl)

    def add_attribute(self, a, enumerations={}):
        """
        Write a definition for attribute a. Enumerations may be a dict
        of enumerations, indexed by ID. These are used to identify enums.
        """
        params = {}
        type = a.typeValue
        if type is None:
            raise ValueError(
                f"ERROR! type is not specified for property {a.class_name}.{a.name}"
            )

        if type.lower() == "boolean":
            type = "int"
        elif type.lower() in ("integer", "unlimitednatural"):
            type = "int"
        elif type.lower() == "string":
            type = "str"

        default = a.defaultValue
        # Make sure types are represented the Python way:
        if default and default.lower() in ("true", "false"):
            default = default.title()  # True or False...
        if default is not None:
            params["default"] = str(default)

        if ast.literal_eval(a.isDerived or "0") and not self.overrides.has_override(
            f"{a.class_name}.{a.name}"
        ):
            msg(f"ignoring derived attribute {a.class_name}.{a.name}): no definition")
        elif type.endswith("Kind") or type.endswith("Sort"):
            e = list(filter(lambda e: e["name"] == type, list(enumerations.values())))[
                0
            ]
            self.add_property(
                a.class_name,
                a.name,
                f"enumeration('{a.name}', {e.enumerates}, '{default or e.enumerates[0]}')",
                type="enumeration",
            )
        else:
            if params:
                attribute = "attribute('{}', {}, {})".format(
                    a.name, type, ", ".join(map("=".join, list(params.items())))
                )
            else:
                attribute = f"attribute('{a.name}', {type})"
            self.add_property(
                a.class_name, a.name, attribute, type=f"attribute[{type}]"
            )

    def add_operation(self, o):
        full_name = f"{o.class_name}.{o.name}"
        if self.overrides.has_override(full_name):
            self.add_property(
                o.class_name,
                o.name,
                self.overrides.get_override(full_name),
                self.overrides.get_type(full_name),
            )
        else:
            msg(f"No override for operation {full_name}")

    def add_association(self, head, tail):
        """
        Write an association for head.
        The association should not be a redefine or derived association.
        """
        if head.written:
            return

        assert head.navigable
        # Derived unions and redefines are handled separately
        assert not head.derived
        assert not head.redefines

        a = f"association('{head.name}', {head.opposite_class_name}"
        if head.lower not in ("0", 0):
            a += f", lower={head.lower}"
        if head.upper != "*":
            a += f", upper={head.upper}"
        if head.composite:
            a += ", composite=True"

        # Add the opposite property if the head itself is navigable:
        if tail.navigable:
            try:
                # o_derived, o_name = parse_association_name(tail['name'])
                o_name = tail.name
                o_derived = tail.derived
            except KeyError:
                msg(
                    f"ERROR! no name, but navigable: {tail.id} ({tail.class_name}.{tail.name})"
                )
            else:
                assert not (
                    head.derived and not o_derived
                ), "One end is derived, the other end not ???"
                a += f", opposite='{o_name}'"

        extension_association_hack = (
            (head.class_name == "Extension" and head.name == "ownedEnd")
            and "  # type: ignore[assignment]"
            or ""
        )

        self.add_property(
            head.class_name,
            head.name,
            a + ")",
            type=f"relation_one[{head.opposite_class_name}]{extension_association_hack}"
            if head.upper == "1"
            else f"relation_many[{head.opposite_class_name}]",
        )

    def add_derivedunion(self, d):
        """
        Write a derived union. If there are no subsets a warning
        is issued. The derivedunion is still created though.

        Derived unions may be created for associations that were returned
        False by add_association().
        """
        subs = ""
        for u in d.union:
            if u.derived and not u.written:
                self.add_derivedunion(u)
            if subs:
                subs += ", "
            subs += f"{u.class_name}.{u.name}"
        if subs:
            self.add_property(
                d.class_name,
                d.name,
                "derivedunion(%s, '%s', %s, %s, %s, %s)"
                % (
                    d.class_name,
                    d.name,
                    d.opposite_class_name,
                    d.lower,
                    d.upper == "*" and "'*'" or d.upper,
                    subs,
                ),
                type=f"relation_one[{d.opposite_class_name}]"
                if d.upper == "1"
                else f"relation_many[{d.opposite_class_name}]",
            )
        else:
            if not self.overrides.has_override(f"{d.class_name}.{d.name}"):
                msg(
                    f"no subsets for derived union: {d.class_name}.{d.name}[{d.lower}..{d.upper}]"
                )
            self.add_property(
                d.class_name,
                d.name,
                "derivedunion(%s, '%s', %s, %s, %s)"
                % (
                    d.class_name,
                    d.name,
                    d.opposite_class_name,
                    d.lower,
                    d.upper == "*" and "'*'" or d.upper,
                ),
                type=f"relation_one[{d.opposite_class_name}]"
                if d.upper == "1"
                else f"relation_many[{d.opposite_class_name}]",
            )
        d.written = True

    def add_redefine(self, r):
        """
        Redefines may be created for associations that were returned
        False by add_association().
        """
        self.add_property(
            r.class_name,
            r.name,
            "redefine(%s, '%s', %s, %s, %s)"
            % (
                r.class_name,
                r.name,
                r.opposite_class_name,
                r.upper == "*" and "'*'" or r.upper,
                r.redefines,
            ),
            type=f"relation_one[{r.opposite_class_name}]  # type: ignore[assignment]"
            if r.upper == "1"
            else f"relation_many[{r.opposite_class_name}]  # type: ignore[assignment]",
        )



"""
Transaction support for Gaphor
"""

import logging
from typing import Callable, List, Set

from gaphor.event import TransactionBegin, TransactionCommit, TransactionRollback

log = logging.getLogger(__name__)

# Sessions should subscribe a hook to receive events emitted via @transactional
subscribers: Set[Callable[[object], None]] = set()


def transactional(func):
    """
    The transactional decorator makes a function transactional.
    Events are emitted through the (global) `subscribers` set.

    It is preferred to use the `Transaction` context manager. The context manager
    emits events in the context of the session in scope, whereas the `@transactional`
    decorator emits a global event which is sent to the active session.
    """

    def _transactional(*args, **kwargs):
        if __debug__ and args and hasattr(args[0], "event_manager"):
            log.warning(f"Consider using the Transaction context manager for {args[0]}")

        with Transaction(_SubscribersHandler):
            return func(*args, **kwargs)

    return _transactional


class TransactionError(Exception):
    """
    Errors related to the transaction module.
    """


class Transaction:
    """
    The transaction. On start and end of a transaction an event is emitted.

    >>> import gaphor.core.eventmanager
    >>> event_manager = gaphor.core.eventmanager.EventManager()

    Transactions can be nested. If the outermost transaction is committed or
    rolled back, an event is emitted.

    Events can be handled programmatically:

    >>> tx = Transaction(event_manager)
    >>> tx.commit()

    It can be assigned as decorator:

    >>> @transactional
    ... def foo():
    ...     pass

    Or with the ``with`` statement:

    >>> with Transaction(event_manager):
    ...     pass
    """

    _stack: List = []

    def __init__(self, event_manager):
        """Initialize the transaction.  If this is the first transaction in
        the stack, a TransactionBegin event is emitted."""
        self.event_manager = event_manager

        self._need_rollback = False
        if not self._stack:
            self._handle(TransactionBegin())
        self._stack.append(self)

    def commit(self):
        """Commit the transaction.  First, the transaction is closed.
        If it needs to be rolled-back, a TransactionRollback event is emitted.
        Otherwise, a TransactionCommit event is emitted."""

        self._close()
        if not self._stack:
            if self._need_rollback:
                self._handle(TransactionRollback())
            else:
                self._handle(TransactionCommit())

    def rollback(self):
        """Roll-back the transaction.  First, the transaction is closed.
        Every transaction on the stack is then marked for roll-back.  If
        the stack is empty, a TransactionRollback event is emitted."""

        self._close()
        for tx in self._stack:
            tx._need_rollback = True
        else:
            if not self._stack:
                self._handle(TransactionRollback())

    def _close(self):
        """Close the transaction.  If the stack is empty, a TransactionError
        is raised.  If the last transaction on the stack isn't this transaction,
        a Transaction error is raised."""

        try:
            last = self._stack.pop()
        except IndexError:
            raise TransactionError("No Transaction on stack.") from None
        if last is not self:
            self._stack.append(last)
            raise TransactionError(
                "Transaction on stack is not the transaction being closed."
            )

    def _handle(self, event):
        self.event_manager.handle(event)

    def __enter__(self):
        """Provide with-statement transaction support."""
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        """Provide with-statement transaction support.  If an error occurred,
        the transaction is rolled back.  Otherwise, it is committed."""

        if exc_type:
            log.error(
                "Transaction terminated due to an exception, performing a rollback",
                exc_info=True,
            )
            try:
                self.rollback()
            except Exception:
                log.error("Rollback failed", exc_info=True)
        else:
            self.commit()


class _SubscribersHandler:
    @classmethod
    def handle(cls, event):
        global subscribers
        for o in subscribers:
            o(event)



import pytest
from gi.repository import Gtk

from gaphor import UML
from gaphor.ui.diagrampage import DiagramPage
from gaphor.UML.modelinglanguage import UMLModelingLanguage
from gaphor.UML.toolbox import uml_toolbox_actions


@pytest.fixture
def properties():
    return {}


@pytest.fixture
def tab(event_manager, element_factory, properties):
    diagram = element_factory.create(UML.Diagram)
    tab = DiagramPage(
        diagram, event_manager, element_factory, properties, UMLModelingLanguage()
    )

    window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
    window.add(tab.construct())
    window.show()
    yield tab
    window.destroy()


def test_pointer(tab):
    tool = tab.get_tool("toolbox-pointer")

    assert tool


@pytest.mark.parametrize(
    "tool_name",
    [
        "toolbox-comment",
        "toolbox-comment-line",
        # Classes:
        "toolbox-class",
        "toolbox-interface",
        "toolbox-package",
        "toolbox-association",
        "toolbox-dependency",
        "toolbox-generalization",
        "toolbox-implementation",
        # Components:
        "toolbox-component",
        "toolbox-node",
        "toolbox-artifact",
        # Actions:
        "toolbox-action",
        "toolbox-initial-node",
        "toolbox-activity-final-node",
        "toolbox-flow-final-node",
        "toolbox-decision-node",
        "toolbox-fork-node",
        "toolbox-flow",
        # Use cases:
        "toolbox-use-case",
        "toolbox-actor",
        "toolbox-use-case-association",
        "toolbox-include",
        "toolbox-extend",
        # Profiles:
        "toolbox-profile",
        "toolbox-metaclass",
        "toolbox-stereotype",
        "toolbox-extension",
    ],
)
def test_placement_action(tab, tool_name):
    print(tool_name)
    tool = tab.get_tool(tool_name)

    # Ensure the factory is working
    tool.create_item((0, 0))
    tab.diagram.canvas.update()


def test_placement_object_node(tab, element_factory):
    test_placement_action(tab, "toolbox-object-node")
    assert len(element_factory.lselect(UML.ObjectNode)) == 1


def test_placement_partition(tab, element_factory):
    test_placement_action(tab, "toolbox-partition")
    assert len(element_factory.lselect(UML.ActivityPartition)) == 0


def test_uml_toolbox_actions_shortcut_unique():

    shortcuts = {}

    for category, items in uml_toolbox_actions:
        for action_name, label, icon_name, shortcut, *rest in items:
            try:
                shortcuts[shortcut].append(action_name)
            except KeyError:
                shortcuts[shortcut] = [action_name]

    for key, val in list(shortcuts.items()):
        if key is not None:
            assert len(val) == 1, f"Duplicate toolbox shortcut {key} for {val}"



import gaphas
import pytest

from gaphor.core.eventmanager import EventManager
from gaphor.core.modeling import ElementFactory, Presentation, StyleSheet
from gaphor.UML import Diagram


class Example(Presentation, gaphas.Element):
    def unlink(self):
        self.test_unlinked = True
        super().unlink()


@pytest.fixture
def element_factory():
    event_manager = EventManager()
    element_factory = ElementFactory()
    yield element_factory
    element_factory.shutdown()
    event_manager.shutdown()


def test_canvas_is_set_up():
    diagram = Diagram("id", None)

    assert diagram.canvas


def test_canvas_is_saved():
    diagram = Diagram("id", None)
    saved_keys = []
    diagram.save(lambda name, val: saved_keys.append(name))

    assert "canvas" in saved_keys


def test_canvas_item_is_created(element_factory):
    diagram = element_factory.create(Diagram)
    example = diagram.create(Example)

    assert example in diagram.canvas.get_all_items()


def test_canvas_is_unlinked(element_factory):
    diagram = element_factory.create(Diagram)
    example = diagram.create(Example)

    diagram.unlink()

    assert example.test_unlinked


def test_can_only_add_diagram_items(element_factory):
    diagram = element_factory.create(Diagram)

    with pytest.raises(TypeError):
        diagram.create(Diagram)


def test_diagram_stylesheet(element_factory):
    diagram = element_factory.create(Diagram)
    styleSheet = element_factory.create(StyleSheet)

    assert diagram.styleSheet is styleSheet


def test_diagram_stylesheet_is_always_there(element_factory):
    diagram = element_factory.create(Diagram)

    assert diagram.styleSheet is not None



"""
This module contains the actions used in the Toolbox (lower left section
of the main window.

The Toolbox is bound to a diagram. When a diagram page (tab) is switched,
the actions bound to the toolbuttons should change as well.
"""

from typing import Callable, NamedTuple, Optional, Sequence, Tuple

from gaphor.core.modeling import Diagram, Presentation

ItemFactory = Callable[[Diagram, Optional[Presentation]], Presentation]


class ToolDef(NamedTuple):
    id: str
    name: str
    icon_name: str
    shortcut: Optional[str]
    item_factory: Optional[ItemFactory]
    handle_index: int = -1


ToolboxDefinition = Sequence[Tuple[str, Sequence[ToolDef]]]



# This file is generated by codegen.py. DO NOT EDIT!

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Callable, List, Optional

from gaphor.core.modeling.element import Element
from gaphor.core.modeling.properties import (
    association,
    attribute,
    derived,
    derivedunion,
    enumeration,
    redefine,
    relation_many,
    relation_one,
)

if TYPE_CHECKING:
    from gaphor.UML import Dependency, Namespace
# 8: override Element
# defined above


# 11: override NamedElement
# Define extra attributes defined in UML model
class NamedElement(Element):
    name: attribute[str]
    qualifiedName: derived[List[str]]
    namespace: relation_one[Namespace]
    clientDependency: relation_many[Dependency]
    supplierDependency: relation_many[Dependency]


class PackageableElement(NamedElement):
    pass


# 55: override Diagram
# defined in gaphor.core.modeling.diagram


# 46: override Presentation
# defined in gaphor.core.modeling.presentation


class Comment(Element):
    body: attribute[str]
    annotatedElement: relation_many[Element]


# 40: override StyleSheet
# defined in gaphor.core.modeling.presentation


NamedElement.name = attribute("name", str)
Comment.body = attribute("body", str)
# 43: override StyleSheet.styleSheet
# defined in gaphor.core.modeling.presentation

# 52: override Presentation.subject
# defined in gaphor.core.modeling.presentation

# 49: override Element.presentation
# defined in gaphor.core.modeling.presentation

Comment.annotatedElement = association(
    "annotatedElement", Element, opposite="ownedComment"
)
Element.ownedComment = association("ownedComment", Comment, opposite="annotatedElement")
# 20: override NamedElement.qualifiedName(NamedElement.namespace): derived[List[str]]


def _namedelement_qualifiedname(self) -> List[str]:
    """
    Returns the qualified name of the element as a tuple
    """
    if self.namespace:
        return _namedelement_qualifiedname(self.namespace) + [self.name]
    else:
        return [self.name]


NamedElement.qualifiedName = derived(
    NamedElement,
    "qualifiedName",
    List[str],
    0,
    1,
    lambda obj: [_namedelement_qualifiedname(obj)],
)



"""
Test include item connections.
"""

from gaphor import UML
from gaphor.tests import TestCase
from gaphor.UML.usecases.include import IncludeItem
from gaphor.UML.usecases.usecase import UseCaseItem


class IncludeItemTestCase(TestCase):
    def test_use_case_glue(self):
        """Test "include" gluing to use cases."""

        uc1 = self.create(UseCaseItem, UML.UseCase)
        include = self.create(IncludeItem)

        glued = self.allow(include, include.head, uc1)
        assert glued

    def test_use_case_connect(self):
        """Test connecting "include" to use cases
        """
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)
        include = self.create(IncludeItem)

        self.connect(include, include.head, uc1)
        assert self.get_connected(include.head), uc1

        self.connect(include, include.tail, uc2)
        assert self.get_connected(include.tail), uc2

    def test_use_case_reconnect(self):
        """Test reconnecting use cases with "include"
        """
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)
        uc3 = self.create(UseCaseItem, UML.UseCase)
        include = self.create(IncludeItem)

        # connect: uc1 -> uc2
        self.connect(include, include.head, uc1)
        self.connect(include, include.tail, uc2)
        e = include.subject

        # reconnect: uc1 -> uc2
        self.connect(include, include.tail, uc3)

        assert e is include.subject
        assert include.subject.addition is uc1.subject
        assert include.subject.includingCase is uc3.subject

    def test_use_case_disconnect(self):
        """Test disconnecting "include" from use cases
        """
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)
        include = self.create(IncludeItem)

        self.connect(include, include.head, uc1)
        self.connect(include, include.tail, uc2)

        self.disconnect(include, include.head)
        assert self.get_connected(include.head) is None
        assert include.subject is None

        self.disconnect(include, include.tail)
        assert self.get_connected(include.tail) is None



from gaphor import UML


def test_qualified_name():
    p1 = UML.Package()
    p2 = UML.Package()
    p3 = UML.Package()

    p1.name = "package1"
    p2.name = "package2"
    p3.name = "package3"
    p3.package = p2
    p2.package = p1

    assert p3.qualifiedName == ["package1", "package2", "package3"]



"""
Connector connections.

Implemented using interface item in assembly connector mode, see
`gaphor.diagram.connector` module for details.
"""

import operator
from typing import Union

from gaphor import UML
from gaphor.diagram.connectors import BaseConnector, Connector
from gaphor.UML.classes.interface import InterfaceItem
from gaphor.UML.components.component import ComponentItem
from gaphor.UML.components.connector import ConnectorItem


@Connector.register(ComponentItem, ConnectorItem)
@Connector.register(InterfaceItem, ConnectorItem)
class ConnectorConnectBase(BaseConnector):
    """
    This connector is left as is, mainly for backwards compatibility.
    The Connector item has been removed from the Components tool
    palette.
    """

    element: Union[ComponentItem, InterfaceItem]
    line: ConnectorItem

    def _get_interfaces(self, c1, c2):
        """
        Return list of common interfaces provided by first component and
        required by second component.

        :Parameters:
         c1
            Component providing interfaces.
         c2
            Component requiring interfaces.
        """
        provided = set(c1.subject.provided)
        required = set(c2.subject.required)
        interfaces = list(provided.intersection(required))
        interfaces.sort(key=operator.attrgetter("name"))
        return interfaces

    def get_connecting(self, iface, both=False):
        """
        Get items connecting to interface.

        :Parameters:
         iface
            Interface in question.
         both
            If true, then filter out one-side connections.
        """
        canvas = iface.canvas
        connected = canvas.get_connections(connected=iface)
        if both:
            connected = [
                c for c in connected if canvas.get_connection(c.item.opposite(c.handle))
            ]
        return connected

    @staticmethod
    def get_component(connector):
        """
        Get component connected by connector.
        """
        canvas = connector.canvas
        c1 = canvas.get_connection(connector.head)
        c2 = canvas.get_connection(connector.tail)
        component = None
        if c1 and isinstance(c1.connected, ComponentItem):
            component = c1.connected
        elif c2 and isinstance(c2.connected, ComponentItem):
            component = c2.connected
        return component

    def create_uml(self, connector, component, assembly, iface):
        """
        Create assembly connector UML metamodel for given connector item
        and component.

        :Parameters:
         connector
            Connector item.
         component
            Component item.
         assembly
            Instance of Connector UML metaclass.
         iface
            Instance of Interface UML metaclass.
        """
        connector.subject = assembly

        end = connector.model.create(UML.ConnectorEnd)
        end.role = iface
        end.partWithPort = connector.model.create(UML.Port)
        assembly.end = end

        component.subject.ownedPort = end.partWithPort

    def drop_uml(self, connector, component):
        """
        Destroy assembly connector UML metamodel existing between connector
        item and component item.

        :Parameters:
         connector
            Connector item.
         component
            Component item.
        """
        if component and component.subject:
            p = component.subject.ownedPort[0]
            p.unlink()
            connector.subject = None

    def allow(self, handle, port):
        iface = self.element
        component = self.get_connected(self.line.opposite(handle))

        if isinstance(component, InterfaceItem) or isinstance(iface, ComponentItem):
            component, iface = iface, component
        # connect only components and interfaces but not two interfaces nor
        # two components
        glue_ok = (not component or isinstance(component, ComponentItem)) and (
            not iface or isinstance(iface, InterfaceItem)
        )

        return glue_ok

    def connect(self, handle, port):
        super().connect(handle, port)

        line = self.line

        c1 = self.get_connected(line.head)
        c2 = self.get_connected(line.tail)
        if c1 and c2:
            # reference interface and component correctly
            iface = c1
            component = c2
            if isinstance(component, InterfaceItem):
                assert isinstance(iface, ComponentItem)
                component, iface = iface, component

            connections = self.get_connecting(iface, both=True)

            assembly = None
            for c in connections:
                if c.item.subject:
                    assembly = c.item.subject
                    assert assembly.kind == "assembly"
                    break

            if line.subject:
                assert isinstance(line.subject, UML.Connector)
                assembly = line.subject
                assert assembly.end
                assert assembly.end[:].partWithPort
                assert iface in assembly.end[:].role
                return

            if assembly is None:
                assembly = self.element.model.create(UML.Connector)
                assembly.kind = "assembly"
                for c in connections:
                    connector = c.item
                    self.create_uml(
                        connector,
                        self.get_component(connector),
                        assembly,
                        iface.subject,
                    )
            else:
                self.create_uml(line, component, assembly, iface.subject)
            iface.request_update()

    def disconnect(self, handle):
        super().disconnect(handle)
        line = self.line
        if line.subject is None:
            return

        iface = self.get_connected(line.head)
        if not isinstance(iface, InterfaceItem):
            iface = self.get_connected(line.tail)

        assert iface, "No interface found on {line}"

        connections = list(self.get_connecting(iface, both=True))

        # destroy whole assembly if one connected item stays
        # or only one port will stay connected
        if len(connections) == 2:
            connector = line.subject
            for ci in connections:
                c = self.get_component(ci.item)
                self.drop_uml(ci.item, c)
                line.request_update(matrix=False)  # type: ignore[call-arg] # noqa: F821
            connector.unlink()
        else:
            c = self.get_component(line)
            self.drop_uml(line, c)
        iface.request_update()



import functools
import importlib
import logging
from typing import Dict, Optional, Sequence, Tuple

from gaphas.freehand import FreeHandPainter
from gaphas.painter import (
    BoundingBoxPainter,
    FocusedItemPainter,
    HandlePainter,
    PainterChain,
    ToolPainter,
)
from gaphas.view import GtkView
from gi.repository import Gdk, GdkPixbuf, GLib, Gtk

from gaphor.core import action, event_handler, gettext, transactional
from gaphor.core.modeling import Presentation, StyleSheet
from gaphor.core.modeling.diagram import StyledDiagram
from gaphor.core.modeling.event import AttributeUpdated, ElementDeleted
from gaphor.diagram.diagramtoolbox import ToolDef
from gaphor.diagram.diagramtools import (
    DefaultTool,
    PlacementTool,
    TransactionalToolChain,
)
from gaphor.diagram.event import DiagramItemPlaced
from gaphor.diagram.painter import ItemPainter
from gaphor.diagram.support import get_diagram_item
from gaphor.transaction import Transaction
from gaphor.ui.actiongroup import create_action_group
from gaphor.ui.event import DiagramSelectionChanged

log = logging.getLogger(__name__)


def tooliter(toolbox_actions: Sequence[Tuple[str, Sequence[ToolDef]]]):
    """
    Iterate toolbox items, regardless of section headers
    """
    for name, section in toolbox_actions:
        yield from section


with importlib.resources.path("gaphor.ui", "placement-icon-base.png") as f:
    PLACEMENT_BASE = GdkPixbuf.Pixbuf.new_from_file_at_scale(str(f), 64, 64, True)

GtkView.set_css_name("diagramview")

_placement_pixbuf_map: Dict[str, GdkPixbuf.Pixbuf] = {}


_upper_offset = ord("A") - ord("a")


@functools.lru_cache(maxsize=None)
def parse_shortcut(shortcut):
    key, mod = Gtk.accelerator_parse(shortcut)
    return (key, key + _upper_offset), mod


def get_placement_cursor(display, icon_name):
    if icon_name in _placement_pixbuf_map:
        pixbuf = _placement_pixbuf_map[icon_name]
    else:
        pixbuf = PLACEMENT_BASE.copy()
        icon = Gtk.IconTheme.get_default().load_icon(icon_name, 24, 0)
        icon.copy_area(
            0, 0, icon.get_width(), icon.get_height(), pixbuf, 9, 15,
        )
        _placement_pixbuf_map[icon_name] = pixbuf
    return Gdk.Cursor.new_from_pixbuf(display, pixbuf, 1, 1)


class DiagramPage:

    VIEW_TARGET_STRING = 0
    VIEW_TARGET_ELEMENT_ID = 1
    VIEW_TARGET_TOOLBOX_ACTION = 2
    VIEW_DND_TARGETS = [
        Gtk.TargetEntry.new("gaphor/element-id", 0, VIEW_TARGET_ELEMENT_ID),
        Gtk.TargetEntry.new("gaphor/toolbox-action", 0, VIEW_TARGET_TOOLBOX_ACTION),
    ]

    def __init__(
        self, diagram, event_manager, element_factory, properties, modeling_language
    ):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.properties = properties
        self.diagram = diagram
        self.modeling_language = modeling_language

        self.view: Optional[GtkView] = None
        self.widget: Optional[Gtk.Widget] = None
        self.diagram_css: Optional[Gtk.CssProvider] = None

        self.event_manager.subscribe(self._on_element_delete)
        self.event_manager.subscribe(self._on_style_sheet_updated)
        self.event_manager.subscribe(self._on_diagram_item_placed)

    title = property(lambda s: s.diagram and s.diagram.name or gettext("<None>"))

    def get_diagram(self):
        return self.diagram

    def get_view(self):
        return self.view

    def construct(self):
        """
        Create the widget.

        Returns: the newly created widget.
        """
        assert self.diagram

        view = GtkView(canvas=self.diagram.canvas)
        view.drag_dest_set(
            Gtk.DestDefaults.ALL,
            DiagramPage.VIEW_DND_TARGETS,
            Gdk.DragAction.MOVE | Gdk.DragAction.COPY | Gdk.DragAction.LINK,
        )
        self.diagram_css = Gtk.CssProvider.new()
        view.get_style_context().add_provider(
            self.diagram_css, Gtk.STYLE_PROVIDER_PRIORITY_USER
        )

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_shadow_type(Gtk.ShadowType.IN)
        scrolled_window.add(view)
        scrolled_window.show_all()
        self.widget = scrolled_window

        view.connect("focus-changed", self._on_view_selection_changed)
        view.connect("selection-changed", self._on_view_selection_changed)
        view.connect_after("key-press-event", self._on_key_press_event)
        view.connect("drag-data-received", self._on_drag_data_received)

        self.view = view

        self.widget.action_group = create_action_group(self, "diagram")

        self.widget.connect_after("key-press-event", self._on_shortcut_action)
        self.select_tool("toolbox-pointer")

        self.set_drawing_style()

        return self.widget

    def get_tool(self, tool_name):
        """
        Return a tool associated with an id (action name).
        """
        if tool_name == "toolbox-pointer":
            return DefaultTool(self.event_manager)

        tool = next(
            t
            for t in tooliter(self.modeling_language.toolbox_definition)
            if t.id == tool_name
        )
        item_factory = tool.item_factory
        handle_index = tool.handle_index
        return PlacementTool(
            self.view,
            item_factory=item_factory,
            event_manager=self.event_manager,
            handle_index=handle_index,
        )

    def get_tool_icon_name(self, tool_name):
        if tool_name == "toolbox-pointer":
            return None
        return next(
            t
            for t in tooliter(self.modeling_language.toolbox_definition)
            if t.id == tool_name
        ).icon_name

    @event_handler(ElementDeleted)
    def _on_element_delete(self, event: ElementDeleted):
        if event.element is self.diagram:
            self.close()

    @event_handler(AttributeUpdated)
    def _on_style_sheet_updated(self, event: AttributeUpdated):
        if event.property is StyleSheet.styleSheet:
            self.set_drawing_style()

            canvas = self.diagram.canvas
            for item in canvas.get_all_items():
                canvas.request_update(item)

    def close(self):
        """
        Tab is destroyed. Do the same thing that would
        be done if Close was pressed.
        """
        assert self.widget
        self.widget.destroy()
        self.event_manager.unsubscribe(self._on_element_delete)
        self.event_manager.unsubscribe(self._on_style_sheet_updated)
        self.event_manager.unsubscribe(self._on_diagram_item_placed)
        self.view = None

    @action(
        name="diagram.zoom-in", shortcut="<Primary>plus",
    )
    def zoom_in(self):
        assert self.view
        self.view.zoom(1.2)

    @action(
        name="diagram.zoom-out", shortcut="<Primary>minus",
    )
    def zoom_out(self):
        assert self.view
        self.view.zoom(1 / 1.2)

    @action(
        name="diagram.zoom-100", shortcut="<Primary>0",
    )
    def zoom_100(self):
        assert self.view
        zx = self.view.matrix[0]
        self.view.zoom(1 / zx)

    @action(
        name="diagram.select-all", shortcut="<Primary>a",
    )
    def select_all(self):
        assert self.view
        if self.view.has_focus():
            self.view.select_all()

    @action(name="diagram.unselect-all", shortcut="<Primary><Shift>a")
    def unselect_all(self):
        assert self.view
        if self.view.has_focus():
            self.view.unselect_all()

    @action(name="diagram.delete")
    @transactional
    def delete_selected_items(self):
        assert self.view
        items = self.view.selected_items
        for i in list(items):
            if isinstance(i, Presentation):
                i.unlink()
            else:
                if i.canvas:
                    i.canvas.remove(i)

    @action(name="diagram.select-tool", state="toolbox-pointer")
    def select_tool(self, tool_name: str):
        if self.view:
            tool = TransactionalToolChain(self.event_manager)
            tool.append(self.get_tool(tool_name))
            self.view.tool = tool
            icon_name = self.get_tool_icon_name(tool_name)
            window = self.view.get_window()
            if icon_name and window:
                window.set_cursor(get_placement_cursor(window.get_display(), icon_name))
            elif window:
                window.set_cursor(None)

    @event_handler(DiagramItemPlaced)
    def _on_diagram_item_placed(self, event):
        assert self.widget
        if self.properties("reset-tool-after-create", True):
            self.widget.action_group.actions.lookup_action("select-tool").activate(
                GLib.Variant.new_string("toolbox-pointer")
            )

    def set_drawing_style(self):
        """
        Set the drawing style for the diagram based on the active style sheet.
        """
        assert self.view
        assert self.diagram_css

        style = self.diagram.style(StyledDiagram(self.diagram, self.view))

        bg = style.get("background-color")
        self.diagram_css.load_from_data(
            f"diagramview {{ background-color: rgba({int(255*bg[0])}, {int(255*bg[1])}, {int(255*bg[2])}, {bg[3]}) }}".encode()
            if bg
            else "".encode()
        )

        sloppiness = style.get("line-style", 0.0)

        item_painter = ItemPainter()

        view = self.view

        if sloppiness:
            item_painter = FreeHandPainter(ItemPainter(), sloppiness=sloppiness)

        view.painter = (
            PainterChain()
            .append(item_painter)
            .append(HandlePainter())
            .append(FocusedItemPainter())
            .append(ToolPainter())
        )
        view.bounding_box_painter = BoundingBoxPainter(item_painter)

        view.queue_draw_refresh()

    def _on_key_press_event(self, view, event):
        """
        Handle the 'Delete' key. This can not be handled directly (through
        GTK's accelerators), otherwise this key will confuse the text
        edit stuff.
        """
        if (
            view.is_focus()
            and event.keyval in (Gdk.KEY_Delete, Gdk.KEY_BackSpace)
            and (
                event.get_state() == 0 or event.get_state() & Gdk.ModifierType.MOD2_MASK
            )
        ):
            self.delete_selected_items()

    def _on_shortcut_action(self, widget, event):
        # accelerator keys are lower case. Since we handle them in a key-press event
        # handler, we'll need the upper-case versions as well in case Shift is pressed.
        for _title, items in self.modeling_language.toolbox_definition:
            for action_name, _label, _icon_name, shortcut, *rest in items:
                if not shortcut:
                    continue
                keys, mod = parse_shortcut(shortcut)
                if event.state == mod and event.keyval in keys:
                    widget.get_toplevel().get_action_group("diagram").lookup_action(
                        "select-tool"
                    ).change_state(GLib.Variant.new_string(action_name))

    def _on_view_selection_changed(self, view, selection_or_focus):
        self.event_manager.handle(
            DiagramSelectionChanged(view, view.focused_item, view.selected_items)
        )

    def _on_drag_data_received(self, view, context, x, y, data, info, time):
        """
        Handle data dropped on the canvas.
        """
        if (
            data
            and data.get_format() == 8
            and info == DiagramPage.VIEW_TARGET_TOOLBOX_ACTION
        ):
            tool = self.get_tool(data.get_data().decode())
            tool.create_item((x, y))
            context.finish(True, False, time)
        elif (
            data
            and data.get_format() == 8
            and info == DiagramPage.VIEW_TARGET_ELEMENT_ID
        ):
            element_id = data.get_data().decode()
            element = self.element_factory.lookup(element_id)
            assert element

            item_class = get_diagram_item(type(element))
            if item_class:
                with Transaction(self.event_manager):
                    item = self.diagram.create(item_class)
                    assert item

                    x, y = view.get_matrix_v2i(item).transform_point(x, y)
                    item.matrix.translate(x, y)
                    item.subject = element

                view.unselect_all()
                view.focused_item = item

            else:
                log.warning(
                    f"No graphical representation for element {type(element).__name__}"
                )
            context.finish(True, False, time)
        else:
            context.finish(False, False, time)



from gaphor import UML
from gaphor.diagram.tests.fixtures import connect, copy_clear_and_paste
from gaphor.UML.classes import AssociationItem, ClassItem, InterfaceItem


def test_class_with_attributes(diagram, element_factory):
    cls = element_factory.create(UML.Class)
    attr = element_factory.create(UML.Property)
    UML.parse(attr, "- attr: str")
    cls.ownedAttribute = attr

    cls_item = diagram.create(ClassItem, subject=cls)

    new_items = copy_clear_and_paste({cls_item}, diagram, element_factory)
    new_cls_item = new_items.pop()

    assert isinstance(new_cls_item, ClassItem)
    assert UML.format(new_cls_item.subject.ownedAttribute[0]) == "- attr: str"


def test_class_with_operation(diagram, element_factory):
    cls = element_factory.create(UML.Class)
    oper = element_factory.create(UML.Operation)
    UML.parse(oper, "- oper(inout param: str): str")
    cls.ownedOperation = oper

    cls_item = diagram.create(ClassItem, subject=cls)

    new_items = copy_clear_and_paste({cls_item}, diagram, element_factory)
    new_cls_item = new_items.pop()

    assert isinstance(new_cls_item, ClassItem)
    assert (
        UML.format(new_cls_item.subject.ownedOperation[0])
        == "- oper(inout param: str): str"
    )


def test_interface_with_attributes_and_operation(diagram, element_factory):
    iface = element_factory.create(UML.Interface)

    attr = element_factory.create(UML.Property)
    UML.parse(attr, "- attr: str")
    iface.ownedAttribute = attr

    oper = element_factory.create(UML.Operation)
    UML.parse(oper, "- oper(inout param: str): str")
    iface.ownedOperation = oper

    iface_item = diagram.create(InterfaceItem, subject=iface)

    new_items = copy_clear_and_paste({iface_item}, diagram, element_factory)
    new_iface_item = new_items.pop()

    assert isinstance(new_iface_item, InterfaceItem)
    assert UML.format(new_iface_item.subject.ownedAttribute[0]) == "- attr: str"
    assert (
        UML.format(new_iface_item.subject.ownedOperation[0])
        == "- oper(inout param: str): str"
    )


def two_classes_and_an_association(diagram, element_factory):
    gen_cls = element_factory.create(UML.Class)
    spc_cls = element_factory.create(UML.Class)
    gen_cls_item = diagram.create(ClassItem, subject=gen_cls)
    spc_cls_item = diagram.create(ClassItem, subject=spc_cls)
    assoc_item = diagram.create(AssociationItem)

    connect(assoc_item, assoc_item.handles()[0], gen_cls_item)
    connect(assoc_item, assoc_item.handles()[1], spc_cls_item)
    UML.model.set_navigability(
        assoc_item.subject, assoc_item.subject.memberEnd[0], True
    )

    return gen_cls_item, spc_cls_item, assoc_item


def test_copy_remove_paste_items_with_connections(diagram, element_factory):
    gen_cls_item, spc_cls_item, assoc_item = two_classes_and_an_association(
        diagram, element_factory
    )

    new_items = copy_clear_and_paste(
        {gen_cls_item, assoc_item, spc_cls_item}, diagram, element_factory
    )
    new_cls1, new_cls2 = element_factory.lselect(UML.Class)
    new_assoc = element_factory.lselect(UML.Association)[0]

    assert new_assoc.memberEnd[0].type in {new_cls1, new_cls2}
    assert new_assoc.memberEnd[1].type in {new_cls1, new_cls2}
    assert new_assoc.memberEnd[0] in new_assoc.memberEnd[1].type.ownedAttribute
    assert new_cls1.presentation[0] in new_items
    assert new_cls2.presentation[0] in new_items
    assert new_assoc.presentation[0] in new_items



"""
This module contains user interface related code, such as the
main screen and diagram windows.
"""

import importlib.resources
import logging
import sys
from typing import Optional

import gi

from gaphor.application import Application
from gaphor.core import event_handler
from gaphor.event import ActiveSessionChanged, SessionShutdown
from gaphor.ui.actiongroup import apply_application_actions

# fmt: off
gi.require_version("Gtk", "3.0")  # noqa: isort:skip
gi.require_version("Gdk", "3.0")  # noqa: isort:skip
from gi.repository import GLib, Gdk, Gio, Gtk  # noqa: isort:skip
# fmt: on


APPLICATION_ID = "org.gaphor.Gaphor"


icon_theme = Gtk.IconTheme.get_default()
with importlib.resources.path("gaphor.ui", "icons") as path:
    icon_theme.append_search_path(str(path))

LOG_FORMAT = "%(name)s %(levelname)s %(message)s"


def main(argv=sys.argv):
    """Start Gaphor from the command line.  This function creates an option
    parser for retrieving arguments and options from the command line.  This
    includes a Gaphor model to load.

    The application is then initialized, passing along the option parser.  This
    provides plugins and services with access to the command line options
    and may add their own."""

    def has_option(*options):
        return any(o in argv for o in options)

    if has_option("-v", "--verbose"):
        logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
    elif has_option("-q", "--quiet"):
        logging.basicConfig(level=logging.WARNING, format=LOG_FORMAT)
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)

    if has_option("-p", "--profiler"):

        import cProfile
        import pstats

        cProfile.runctx(
            "run(argv)", globals(), locals(), filename="gaphor.prof",
        )

        profile_stats = pstats.Stats("gaphor.prof")
        profile_stats.strip_dirs().sort_stats("time").print_stats(50)

    else:
        run(argv)


def subscribe_to_lifecycle_events(session, application, gtk_app):
    @event_handler(ActiveSessionChanged)
    def on_active_session_changed(event):
        application.active_session = session

    @event_handler(SessionShutdown)
    def on_session_shutdown(event):
        application.shutdown_session(session)
        if not application.sessions:
            gtk_app.quit()

    event_manager = session.get_service("event_manager")
    event_manager.subscribe(on_active_session_changed)
    event_manager.subscribe(on_session_shutdown)


def run(args):
    application: Optional[Application] = None

    def new_session():
        assert application
        session = application.new_session()
        subscribe_to_lifecycle_events(session, application, gtk_app)

        main_window = session.get_service("main_window")
        main_window.open(gtk_app)

        return session

    def app_startup(gtk_app):
        nonlocal application
        try:
            application = Application()
            apply_application_actions(application, gtk_app)
        except Exception:
            gtk_app.quit()
            raise

    def app_activate(gtk_app):
        assert application
        if not application.has_sessions():
            gtk_app.open([], "__new__")

    def app_open(gtk_app, files, n_files, hint):
        session = new_session()
        file_manager = session.get_service("file_manager")
        if hint == "__new__":
            file_manager.new()
        else:
            assert n_files == 1
            for file in files:
                file_manager.load(file.get_path())

    gtk_app = Gtk.Application(
        application_id=APPLICATION_ID, flags=Gio.ApplicationFlags.HANDLES_OPEN
    )
    add_main_options(gtk_app)
    gtk_app.connect("startup", app_startup)
    gtk_app.connect("activate", app_activate)
    gtk_app.connect("open", app_open)
    gtk_app.run(args)


def add_main_options(gtk_app):
    """
    These parameters are handled in `gaphor.ui.main()`.
    Define them here, so they show up on `gaphor --help`.
    """
    gtk_app.add_main_option(
        "verbose",
        ord("v"),
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Verbose output",
        None,
    )
    gtk_app.add_main_option(
        "quiet",
        ord("q"),
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Quiet output",
        None,
    )
    gtk_app.add_main_option(
        "profiler",
        ord("p"),
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Run in profiler",
        None,
    )



from __future__ import annotations

from gaphor.core.modeling import Element
from gaphor.core.modeling.event import AttributeUpdated
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import CompiledStyleSheet, Style, StyleNode


class StyleSheet(Element):
    _compiled_style_sheet: CompiledStyleSheet

    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self.compile_style_sheet()

    styleSheet: attribute[str] = attribute("styleSheet", str, "")

    def compile_style_sheet(self) -> None:
        self._compiled_style_sheet = CompiledStyleSheet(self.styleSheet)

    def match(self, node: StyleNode) -> Style:
        return self._compiled_style_sheet.match(node)

    def postload(self):
        super().postload()
        self.compile_style_sheet()

    def handle(self, event):
        # Ensure compiled style sheet is always up to date:
        if (
            isinstance(event, AttributeUpdated)
            and event.property is StyleSheet.styleSheet
        ):
            self.compile_style_sheet()

        super().handle(event)



"""
Node item may represent a node or a device UML metamodel classes.

Grouping
========
Node item can group following items

- other nodes, which are represented with Node.nestedNode on UML metamodel
  level
- deployed artifacts using deployment
- components, which are parts of a node acting as structured classifier
  (nodes may have internal structures)

Node item grouping logic is implemented in `gaphor.adapters.grouping`
module.
"""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import Classified, ElementPresentation
from gaphor.diagram.shapes import (
    Box,
    EditableText,
    Text,
    VerticalAlign,
    draw_highlight,
    stroke,
)
from gaphor.diagram.support import represents
from gaphor.diagram.text import FontWeight
from gaphor.UML.classes.stereotype import stereotype_compartments


@represents(UML.Node)
@represents(UML.Device)
class NodeItem(ElementPresentation, Classified):
    """
    Representation of node or device from UML Deployment package.
    """

    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self.watch("show_stereotypes", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype", self.update_shapes)
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject.appliedStereotype.slot", self.update_shapes)
        self.watch("subject.appliedStereotype.slot.definingFeature.name")
        self.watch("subject.appliedStereotype.slot.value", self.update_shapes)
        self.watch("subject[Node].ownedConnector", self.update_shapes)
        self.watch("subject[Node].deployment", self.update_shapes)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    def update_shapes(self, event=None):
        self.shape = Box(
            Box(
                Text(
                    text=lambda: UML.model.stereotypes_str(
                        self.subject,
                        isinstance(self.subject, UML.Device) and ("device",) or (),
                    ),
                ),
                EditableText(
                    text=lambda: self.subject.name or "",
                    style={"font-weight": FontWeight.BOLD},
                ),
                style={"padding": (4, 4, 4, 4)},
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            style={
                "min-width": 100,
                "min-height": 50,
                "vertical-align": VerticalAlign.TOP
                if self.canvas and self.canvas.get_children(self)
                else VerticalAlign.MIDDLE,
            },
            draw=draw_node
        )

    def postload(self):
        self.update_shapes()
        super().postload()


def draw_node(box, context, bounding_box):
    cr = context.cairo

    d = 10
    w = bounding_box.width
    h = bounding_box.height

    cr.rectangle(0, 0, w, h)

    draw_highlight(context)

    cr.move_to(0, 0)
    cr.line_to(d, -d)
    cr.line_to(w + d, -d)
    cr.line_to(w + d, h - d)
    cr.line_to(w, h)
    cr.move_to(w, 0)
    cr.line_to(w + d, -d)

    stroke(context)



"""
Test if the collection's list supports all trickery.
"""

import pytest

from gaphor.core.modeling.collection import collection, collectionlist


class MockElement:
    def __init__(self):
        self.events = []

    def handle(self, event):
        self.events.append(event)


class MockProperty:
    def __init__(self):
        self.values = []

    def _set(self, object, value):
        self.values.append((object, value))


def test_listing():
    c: collectionlist[str] = collectionlist()
    c.append("a")
    c.append("b")
    c.append("c")
    assert str(c) == "['a', 'b', 'c']"


def test_append():
    p = MockProperty()
    o = object()
    c = collection(p, o, str)
    c.append("s")

    assert p.values == [(o, "s")]


def test_append_wrong_type():
    c = collection(None, None, int)

    with pytest.raises(TypeError):
        c.append("s")  # type: ignore[arg-type]


def test_size():
    c = collection(None, None, int)
    c.items = [1, 2]  # type: ignore[assignment]

    assert c.size() == 2


def test_includes():
    c = collection(None, None, int)
    c.items = [1, 2]  # type: ignore[assignment]

    assert c.includes(1)
    assert not c.includes(3)


def test_excludes():
    c = collection(None, None, int)
    c.items = [1, 2]  # type: ignore[assignment]

    assert not c.excludes(1)
    assert c.excludes(3)


def test_count():
    c = collection(None, None, int)
    c.items = [1, 2, 2]  # type: ignore[assignment]

    assert c.count(1) == 1
    assert c.count(2) == 2
    assert c.count(3) == 0


def test_includesAll():
    c = collection(None, None, int)
    c.items = [1, 2]  # type: ignore[assignment]

    assert c.includesAll([1])
    assert not c.includesAll([3])


def test_excludesAll():
    c = collection(None, None, int)
    c.items = [1, 2]  # type: ignore[assignment]

    assert not c.excludesAll([1])
    assert c.excludesAll([3])


def test_select():
    c = collection(None, None, int)
    c.items = [1, 2]  # type: ignore[assignment]

    assert c.select(lambda e: e > 1) == [2]


def test_reject():
    c = collection(None, None, int)
    c.items = [1, 2]  # type: ignore[assignment]

    assert c.reject(lambda e: e > 1) == [1]


def test_collect():
    c = collection(None, None, int)
    c.items = [1, 2, 3]  # type: ignore[assignment]

    assert c.collect(lambda e: e * e) == [1, 4, 9]


def test_empty():
    c = collection(None, None, int)

    assert c.isEmpty()


def test_not_empty():
    c: collection[int] = collection(None, None, int)
    c.items = [1, 2, 3]  # type: ignore[assignment]

    assert not c.isEmpty()


def test_swap():
    o = MockElement()
    c: collection[str] = collection(None, o, str)
    c.items = ["a", "b", "c"]  # type: ignore[assignment]
    c.swap("a", "c")
    assert c.items == ["c", "b", "a"]
    assert o.events
